from typing import List, Optional

import pymongo

from statina.adapter import StatinaAdapter
from statina.crud.find import find


def get_last_batches(adapter, nr_of_batches: int) -> list:
    """getting the <nr_of_batches> last batches based on SequencingDate"""

    batches: pymongo.cursor = adapter.batch_collection.find().sort([("SequencingDate", -1)])
    return list(batches.limit(nr_of_batches))


def get_statistics_for_scatter_plot(batches: list, fields: list) -> dict:
    """Formatting data for scatter plot"""

    scatter_plot_data = {}
    for batch in batches:
        batch_id = batch.get("batch_id")
        scatter_plot_data[batch_id] = {"date": batch.get("SequencingDate")}

        for field in fields:
            scatter_plot_data[batch_id][field] = batch.get(field)

    return scatter_plot_data


def get_statistics_for_box_plot(adapter: StatinaAdapter, batches: list, fields: list):
    """Getting and formatting data for box plot"""

    match = {"$match": {"batch_id": {"$in": batches}}}
    lookup = {
        "$lookup": {
            "from": "batch",
            "localField": "batch_id",
            "foreignField": "batch_id",
            "as": "batch",
        }
    }
    unwind = {"$unwind": {"path": "$batch"}}
    group = {
        "$group": {
            "_id": {"batch": "$batch_id", "date": "$batch.SequencingDate"},
            "sample_ids": {"$push": "$sample_id"},
        },
    }

    for field in fields:
        group["$group"][field] = {"$push": f"${field}"}

    pipe = [match, lookup, unwind, group]
    # maybe add a fina sort to the pipe
    box_plot_data = list(find.sample_aggregate(pipe=pipe, adapter=adapter))
    return {batch["_id"]["batch"]: batch for batch in box_plot_data}



import datetime
from typing import Optional

from fastapi import APIRouter, Depends
from fastapi.responses import RedirectResponse
from fastapi.security import OAuth2PasswordRequestForm

from statina.API.external.api.deps import authenticate_user, create_access_token
from statina.config import settings
from statina.models.database import User

router = APIRouter()


@router.post("/token")
def login_for_access_token(form_data: OAuth2PasswordRequestForm = Depends()) -> Optional[str]:
    """Creating a time delimited access token if the user is found in the database."""

    user: User = authenticate_user(form_data.username, form_data.password)

    if not user:
        return None

    access_token_expires = datetime.timedelta(minutes=settings.access_token_expire_minutes)

    return create_access_token(
        username=user.username,
        form_data=form_data,
        expires_delta=access_token_expires,
    )


@router.get("/logout")
def logout():
    """Drop token from cookie and redirects back to index"""

    response = RedirectResponse("../")
    response.set_cookie(key="token", value="")
    return response


@router.post("/login")
def login(token: Optional[str] = Depends(login_for_access_token)):
    """Redirects back to index, if invalid username or password"""

    if not token:
        response = RedirectResponse("../")
        response.set_cookie(key="info_type", value="danger")
        response.set_cookie(
            key="user_info",
            value=f"Wrong username or password.",
        )
    else:
        response = RedirectResponse("../batches")
        response.set_cookie(key="token", value=token)
    return response



import datetime
import smtplib
import ssl
from email.message import EmailMessage
from typing import List

from fastapi import Depends
from pydantic import EmailStr
from starlette.requests import Request
from starlette.responses import RedirectResponse

from statina.API.external.api.api_v1.endpoints.login import router
from statina.API.external.api.deps import get_password_hash, get_current_user
from statina.adapter import StatinaAdapter
from statina.config import get_nipt_adapter, templates
from statina.crud.find import find
from statina.crud.insert import insert_user
from statina.exeptions import EmailNotSentError, CredentialsError
from statina.models.database import User
from statina.models.server.new_user import NewUserRequestEmail, NewUser


def send_mail(user: str, email: EmailStr):
    """Email handler. Sending new user request email to admin"""

    try:
        email_info = NewUserRequestEmail(message=user, user_email=email)
        context = ssl.create_default_context()
        with smtplib.SMTP_SSL(
            email_info.smtp_server, email_info.sll_port, context=context
        ) as server:
            server.login(email_info.sender_email, email_info.sender_password)
            msg = EmailMessage()
            msg.set_content(email_info.message)
            msg["Subject"] = email_info.subject
            msg["From"] = email_info.sender_email
            msg["To"] = email_info.admin_email
            server.send_message(msg)
    except:
        raise EmailNotSentError(
            message=f"Your user has been created, but not activated. Please send an email to "
            f"admin and ask for permission."
        )


@router.post("/add_new_user")
async def add_new_user(request: Request, adapter: StatinaAdapter = Depends(get_nipt_adapter)):
    """Redirects back to index, if invalid username or password"""
    form = await request.form()
    new_user = NewUser(**form)

    user = User(
        **new_user.dict(),
        added=datetime.datetime.now(),
        role="inactive",
        hashed_password=get_password_hash(new_user.password),
    )

    response = RedirectResponse("new_user")

    try:
        insert_user(adapter=adapter, user=user)
        send_mail(
            user=new_user.username, email=new_user.email
        )  # if thisone fails, the error is not picked up!?!?!
        response.set_cookie(key="info_type", value="success")
        response.set_cookie(
            key="user_info",
            value=f"Your user account has been created and an email has been sent to the Statina admin. "
            f"They will send an email to {new_user.email} when your user has been confirmed and activated.",
        )

    except Exception as error:
        response.set_cookie(key="info_type", value="danger")
        response.set_cookie(key="user_info", value=f"{error}")
        pass

    return response


@router.get("/new_user")
def new_user(request: Request):
    """Log in view."""
    return templates.TemplateResponse(
        "new_user.html", context={"request": request, "current_user": ""}
    )


@router.post("/new_user")
def new_user(request: Request):
    """Log in view."""
    return templates.TemplateResponse(
        "new_user.html",
        context={
            "request": request,
            "current_user": "",
            "info_type": request.cookies.get("info_type"),
            "user_info": request.cookies.get("user_info"),
        },
    )


@router.get("/users")
def users(
    request: Request,
    adapter: StatinaAdapter = Depends(get_nipt_adapter),
    user: User = Depends(get_current_user),
):
    """Admin view with table of all users."""
    if user.role != "admin":
        raise CredentialsError(message="Only admin users can access the users page")

    user_list: List[User] = find.users(adapter=adapter)
    return templates.TemplateResponse(
        "users.html",
        context={
            "request": request,
            "users": user_list,
            "page_id": "users",
            "current_user": user,
        },
    )


@router.post("/users")
def users(
    request: Request,
    adapter: StatinaAdapter = Depends(get_nipt_adapter),
    user: User = Depends(get_current_user),
):
    """Admin view with table of all users."""
    if user.role != "admin":
        CredentialsError(message="Only admin users can access the users page")

    user_list: List[User] = find.users(adapter=adapter)
    return templates.TemplateResponse(
        "users.html",
        context={
            "request": request,
            "users": user_list,
            "page_id": "users",
            "current_user": user,
        },
    )



from pathlib import Path

from fastapi.templating import Jinja2Templates
from pydantic import BaseSettings, EmailStr
from pymongo import MongoClient

from statina.adapter.plugin import StatinaAdapter

NIPT_PACKAGE = Path(__file__).parent
PACKAGE_ROOT: Path = NIPT_PACKAGE.parent
ENV_FILE: Path = PACKAGE_ROOT / ".env"
TEMPLATES_DIR: Path = NIPT_PACKAGE / "API" / "external" / "api" / "api_v1" / "templates"


class Settings(BaseSettings):
    """Settings for serving the statina app and connect to the mongo database"""

    db_uri: str = "mongodb://localhost:27017/statina-demo"
    db_name: str = "statina-demo"
    secret_key: str = "dummy"
    algorithm: str = "ABC"
    host: str = "localhost"
    access_token_expire_minutes: int = 60
    port: int = 8000

    class Config:
        env_file = str(ENV_FILE)


class EmailSettings(BaseSettings):
    """Settings for sending email"""

    sll_port: int = 465
    smtp_server: str = "smtp.gmail.com"
    sender_email: EmailStr
    sender_password: str
    admin_email: EmailStr

    class Config:
        env_file = str(ENV_FILE)


settings = Settings()
templates = Jinja2Templates(directory=str(TEMPLATES_DIR))


def get_nipt_adapter():
    client = MongoClient(settings.db_uri)
    return StatinaAdapter(client, db_name=settings.db_name)



from datetime import datetime, timedelta
from typing import Optional

from fastapi import Depends, Request
from fastapi.security import OAuth2PasswordRequestForm
from jose import JWTError, jwt
from passlib.context import CryptContext

from statina.adapter.plugin import StatinaAdapter
from statina.config import get_nipt_adapter, settings
from statina.crud.find import find
from statina.exeptions import CredentialsError
from statina.models.database import User

pwd_context = CryptContext(schemes=["bcrypt"], deprecated="auto")


def get_current_user(
    request: Request,
    adapter: StatinaAdapter = Depends(get_nipt_adapter),
) -> User:
    """Decoding user from token stored in cookies."""
    try:
        payload = jwt.decode(
            request.cookies.get("token"),
            settings.secret_key,
            algorithms=[settings.algorithm],
        )
        username: str = payload.get("sub")
        if username is None:
            raise CredentialsError(message="Could not validate credentials")
    except JWTError:
        raise CredentialsError(message="Could not validate credentials")
    user: User = find.user(adapter=adapter, user_name=username)
    if not user:
        raise CredentialsError(message="User not found in database.")
    return user


def verify_password(plain_password: str, hashed_password: str) -> bool:
    return pwd_context.verify(plain_password, hashed_password)


def get_password_hash(password: str) -> str:
    return pwd_context.hash(password)


def authenticate_user(username: str, password: str) -> Optional[User]:
    adapter: StatinaAdapter = get_nipt_adapter()
    user: User = find.user(adapter=adapter, user_name=username)

    if not user:
        return None
    if user.role == "inactive":
        return None
    if not verify_password(password, user.hashed_password):
        return None
    return user


def create_access_token(
    username: str, form_data: OAuth2PasswordRequestForm, expires_delta: Optional[timedelta] = None
) -> str:
    to_encode = {"sub": username, "scopes": form_data.scopes}
    if expires_delta:
        expire = datetime.utcnow() + expires_delta
    else:
        expire = datetime.utcnow() + timedelta(minutes=15)
    to_encode.update({"exp": expire})
    return jwt.encode(to_encode, settings.secret_key, algorithm=settings.algorithm)



from typing import Optional

from fastapi import status


class NIPToolError(Exception):
    def __init__(self, message: str):
        self.message = message


class MissingResultsError(NIPToolError):
    pass


class InvalidFileError(NIPToolError):
    pass


class FileValidationError(NIPToolError):
    pass


class MissMatchingPasswordError(NIPToolError):
    """Raise when password and confirmed password dont match"""

    pass


class EmailNotSentError(NIPToolError):
    pass


class NIPToolRestError(NIPToolError):
    def __init__(self, message: str, code: Optional[int] = None):
        self.message = message
        self.code = code
        super().__init__(message)


class InsertError(NIPToolRestError):
    def __init__(self, message: str, code: Optional[int] = status.HTTP_405_METHOD_NOT_ALLOWED):
        self.message = message
        self.code = code
        super().__init__(message, code)


class CredentialsError(NIPToolError):
    def __init__(self, message: str, code: Optional[int] = status.HTTP_401_UNAUTHORIZED):
        self.message = message
        self.code = code
        super().__init__(message)



from statina.adapter import StatinaAdapter
from statina.crud.find.plots.statistics_plot_data import (
    get_last_batches,
    get_statistics_for_box_plot,
    get_statistics_for_scatter_plot,
)


def test_get_last_batches(database):
    # GIVEN a database with four batch documents:
    nipt_adapter = StatinaAdapter(database.client, db_name="test")

    batch1 = {"SequencingDate": "2022-03-10"}
    batch2 = {"SequencingDate": "2022-02-10"}
    batch3 = {"SequencingDate": "2022-02-09"}
    batch4 = {"SequencingDate": "2021-03-10"}

    nipt_adapter.batch_collection.insert_many([batch4, batch1, batch3, batch2])

    # WHEN running get_last_batches with nr=3
    results = get_last_batches(adapter=nipt_adapter, nr_of_batches=3)

    # THEN the results should contain the thre batches with the latest
    # SequencingDate, sorted by SequencingDate
    assert results == [batch1, batch2, batch3]


def test_get_last_batches_to_fiew_batches_in_database(database):
    # GIVEN a database with four batch documents:
    nipt_adapter = StatinaAdapter(database.client, db_name="test")

    batch1 = {"SequencingDate": "2022-03-10"}
    batch2 = {"SequencingDate": "2022-02-10"}
    batch3 = {"SequencingDate": "2022-02-09"}
    batch4 = {"SequencingDate": "2021-03-10"}

    nipt_adapter.batch_collection.insert_many([batch4, batch1, batch3, batch2])

    # WHEN running get_last_batches with nr=5
    results = get_last_batches(adapter=nipt_adapter, nr_of_batches=5)

    # THEN the results should contain the four batches in the
    # database sorted by SequencingDate
    assert results == [batch1, batch2, batch3, batch4]


def test_get_statistics_for_box_plot(database, small_helpers):
    # GIVEN a database with thre batches and nine samples like below
    nipt_adapter = StatinaAdapter(database.client, db_name="test")
    batches = ["201860", "201861", "201862"]
    for batch_id in batches:
        batch = small_helpers.batch(batch_id=batch_id)
        nipt_adapter.batch_collection.insert_one(batch)
        for i in [1, 2, 3]:
            sample = small_helpers.sample(
                sample_id=batch_id + str(i), batch_id=batch_id, ratio_13=i, fetal_fraction=i
            )
            nipt_adapter.sample_collection.insert_one(sample)

    # WHEN running get_statistics_for_box_plot
    results = get_statistics_for_box_plot(
        adapter=nipt_adapter,
        batches=batches,
        fields=["Ratio_13", "FF_Formatted"],
    )

    # THEN the results shoule be:
    for batch_id in batches:
        assert results[batch_id] == {
            "FF_Formatted": [1, 2, 3],
            "Ratio_13": [1, 2, 3],
            "_id": {"batch": batch_id, "date": "2022-03-10"},
        }


def test_get_statistics_for_scatter_plot(small_helpers):
    # GIVEN a list with three batch documents:
    batches = [
        small_helpers.batch(batch_id="201860", stdev_13=0.02, stdev_21=0.01),
        small_helpers.batch(batch_id="201861", stdev_13=0.01, stdev_21=0.09),
        small_helpers.batch(batch_id="201862", stdev_13=0.05, stdev_21=0.08),
    ]

    # WHEN running get_statistics_for_scatter_plot
    results = get_statistics_for_scatter_plot(batches=batches, fields=["Stdev_13", "Stdev_21"])

    # THEN the results should look like this:
    expected_result = {
        "201860": {"date": "2022-03-10", "Stdev_13": 0.02, "Stdev_21": 0.01},
        "201861": {"date": "2022-03-10", "Stdev_13": 0.01, "Stdev_21": 0.09},
        "201862": {"date": "2022-03-10", "Stdev_13": 0.05, "Stdev_21": 0.08},
    }
    assert results == expected_result



from datetime import datetime
from pathlib import Path
from typing import List

from fastapi import APIRouter, Depends, Response, status

from statina.adapter.plugin import StatinaAdapter
from statina.API.external.api.deps import get_password_hash
from statina.config import get_nipt_adapter
from statina.crud.find import find
from statina.crud.insert import insert_batch, insert_samples, insert_user
from statina.models.database import Batch, DataBaseSample, User
from statina.models.server.load import BatchRequestBody, UserRequestBody
from statina.parse.batch import get_batch, get_samples

router = APIRouter()


@router.post("/batch")
def batch(
    response: Response,
    batch_files: BatchRequestBody,
    adapter: StatinaAdapter = Depends(get_nipt_adapter),
):
    """Function to load batch data into the database with rest"""
    nipt_results = Path(batch_files.result_file)
    if not nipt_results.exists():
        response.status_code = status.HTTP_422_UNPROCESSABLE_ENTITY
        return {"message": "Results file missing."}
    samples: List[DataBaseSample] = get_samples(nipt_results)
    batch: Batch = get_batch(nipt_results)
    if find.batch(adapter=adapter, batch_id=batch.batch_id):
        return "batch already in database"

    insert_batch(adapter=adapter, batch=batch, batch_files=batch_files)
    insert_samples(adapter=adapter, samples=samples, segmental_calls=batch_files.segmental_calls)

    response.status_code = status.HTTP_200_OK
    return {"message": f"Batch {batch.batch_id} inserted to the database"}


@router.post("/user")
def user(
    response: Response, user: UserRequestBody, adapter: StatinaAdapter = Depends(get_nipt_adapter)
):
    """Function to load user into the database with rest"""

    user_dict = {
        "email": user.email,
        "username": user.username,
        "role": user.role,
        "hashed_password": get_password_hash(user.password),
        "added": datetime.now(),
    }

    if find.user(adapter=adapter, email=user.email):
        return "user already in database"

    insert_user(adapter=adapter, user=User(**user_dict))

    response.status_code = status.HTTP_200_OK
    return {"message": f"User {user.email} inserted to the database."}



from pathlib import Path

import pytest
from fastapi.testclient import TestClient
from mongomock import MongoClient

from statina.adapter.plugin import StatinaAdapter
from statina.config import get_nipt_adapter
from statina.main import internal_app as app
from statina.models.server.load import BatchRequestBody, UserRequestBody

from .small_helpers import SmallHelpers

DATABASE = "testdb"


def override_nipt_adapter():
    """Function for overriding the statina adapter dependency"""

    mongo_client = MongoClient()
    database = mongo_client[DATABASE]
    return StatinaAdapter(database.client, db_name=DATABASE)


@pytest.fixture()
def fast_app_client():
    """Return a mock fastapi app"""

    client = TestClient(app)
    app.dependency_overrides[get_nipt_adapter] = override_nipt_adapter
    return client


#####################


@pytest.fixture(scope="function")
def pymongo_client(request):
    """Get a client to the mongo database"""

    mock_client = MongoClient()

    def teardown():
        mock_client.drop_database(DATABASE)

    request.addfinalizer(teardown)
    return mock_client


@pytest.fixture(scope="function")
def database(request, pymongo_client):
    """Get an adapter connected to mongo database"""

    mongo_client = pymongo_client
    database = mongo_client[DATABASE]
    return database


@pytest.fixture(scope="function")
def valid_load_user():
    return UserRequestBody(
        email="maya.papaya@something.se",
        username="Maya Papaya",
        role="RW",
        password="123",
    )


@pytest.fixture(scope="function")
def valid_load_batch(multiqc_report, segmental_calls, valid_csv):
    batch_files = BatchRequestBody(
        multiqc_report=multiqc_report, segmental_calls=segmental_calls, result_file=valid_csv
    )
    return batch_files


@pytest.fixture(name="small_helpers")
def fixture_small_helpers():
    """Return a class with small helper functions"""
    return SmallHelpers()


##########################################
###### fixture files for input csv ######
##########################################


@pytest.fixture
def valid_csv() -> Path:
    """Get file path to valid csv"""

    return "tests/fixtures/fluffy_result_files/valid_fluffy.csv"


@pytest.fixture
def csv_with_missing_sample_id():
    """Get file path to invalid csv"""

    return "tests/fixtures/fluffy_result_files/fluffy_file_with_missing_sample_id.csv"


@pytest.fixture
def csv_with_missing_sample_project():
    """Get file path to invalid csv"""

    return "tests/fixtures/fluffy_result_files/fluffy_file_with_missing_sample_project.csv"


@pytest.fixture
def multiqc_report():
    """Get file path to multiqc_report"""

    return "tests/fixtures/fluffy_result_files/multiqc_report.html"


@pytest.fixture
def segmental_calls():
    """Get file path to segmental_calls"""

    return "tests/fixtures/fluffy_result_files"



from typing import List, Optional

from pydantic import BaseModel, Field, validator


class ZscoreSamples(BaseModel):
    """Data points for Zscore plots.

    Samples within this model will be part of the same series in a plot.

    The x-axis is of type int because:
        Zscore plots for a batch have samples on the x-axis.
        Zscore plot for a sample have chromosome abnormalities (13 18 21)"""

    count: Optional[int]
    x_axis: Optional[List[int]]
    names: List[str]
    ncv_values: List[float]

    @validator("ncv_values")
    def check_list_lengths(cls, v, values: dict) -> List[int]:

        if len(v) != len(values["names"]):
            raise ValueError("Axis lengths do not match 1")

        if values["x_axis"] is not None and len(v) != len(values["x_axis"]):
            raise ValueError("Axis lengths do not match 2")

        if values["count"] is not None and len(v) != values["count"]:
            raise ValueError("Axis lengths do not match 3")
        return v

    class Config:
        validate_assignment = True


class Zscore131821(BaseModel):
    """Model for samples classified as normal chromosome 13, 18, 21.

    Samples are grouped by chromosome."""

    chr_13: ZscoreSamples = Field(..., alias="13")
    chr_18: ZscoreSamples = Field(..., alias="18")
    chr_21: ZscoreSamples = Field(..., alias="21")

    class Config:
        allow_population_by_field_name = True



from typing import Optional

from pydantic import BaseModel, EmailStr
from typing_extensions import Literal


class BatchRequestBody(BaseModel):
    result_file: str
    multiqc_report: Optional[str]
    segmental_calls: Optional[str]


class UserRequestBody(BaseModel):
    email: EmailStr
    username: str
    password: str
    role: Literal["R", "RW", "admin"]



from typing import List, Optional

from pydantic import BaseModel, validator


class FetalFractionSamples(BaseModel):
    """Data points for fetal fraction plots.

    Samples within this model will be part of the same series in a plot."""

    FFY: List[float]
    FFX: List[float]
    FF: Optional[List[float]]
    names: List[str]
    count: int

    @validator("count")
    def check_list_lengths(cls, v, values: dict) -> List[int]:
        for key, plot_list in values.items():
            if plot_list is not None and v != len(plot_list):
                raise ValueError("Axis lengths do not match")
        return v


class Suspected(FetalFractionSamples):
    """Color code for suspected sample"""

    color: Optional[str] = "#DBA901"
    color_group: Optional[str] = "warning"


class Probable(FetalFractionSamples):
    """Color code for probable sample"""

    color: Optional[str] = "#0000FF"
    color_group: Optional[str] = "warning"


class FalseNegative(FetalFractionSamples):
    """Color code for false negative sample"""

    color: Optional[str] = "#ff6699"
    color_group: Optional[str] = "danger"


class Verified(FetalFractionSamples):
    """Color code for verified sample"""

    color: Optional[str] = "#00CC00"
    color_group: Optional[str] = "danger"


class Other(FetalFractionSamples):
    """Color code for other sample"""

    color: Optional[str] = "#603116"
    color_group: Optional[str] = "warning"


class FalsePositive(FetalFractionSamples):
    """Color code for false positive sample"""

    color: Optional[str] = "#E74C3C"
    color_group: Optional[str] = "success"


class Failed(FetalFractionSamples):
    """Color code for failed sample"""

    color_group: Optional[str] = "danger"


class AbNormalityClasses(BaseModel):
    """A sample can be classified with eg probable X0. False Negative XXY, etc.

    This model is for grouping a set of samples by classification of a specific abnormality, eg XXY.

    All classes are optional since, eg, it can be the case that none of the samples in the set are classified as
    False Positive XXY.

    The status_data_ is a dict of all samples in the set, grouped by classification.
    """

    status_data_: dict
    probable: Optional[Probable]
    suspected: Optional[Suspected]
    false_positive: Optional[FalsePositive]
    verified: Optional[Verified]
    false_negative: Optional[FalseNegative]
    other: Optional[Other]
    failed: Optional[Failed]

    @validator("probable", always=True)
    def set_probable(cls, v, values: dict) -> Optional[Probable]:
        status_info: Optional[dict] = values["status_data_"].get("Probable")
        if not status_info:
            return None
        return Probable(**status_info)

    @validator("suspected", always=True)
    def set_suspected(cls, v, values: dict) -> Optional[Suspected]:
        status_info: Optional[dict] = values["status_data_"].get("Suspected")
        if not status_info:
            return None
        return Suspected(**status_info)

    @validator("false_positive", always=True)
    def set_false_positive(cls, v, values: dict) -> Optional[FalsePositive]:
        status_info: Optional[dict] = values["status_data_"].get("False Positive")
        if not status_info:
            return None
        return FalsePositive(**status_info)

    @validator("verified", always=True)
    def set_verified(cls, v, values: dict) -> Optional[Verified]:
        status_info: Optional[dict] = values["status_data_"].get("Verified")
        if not status_info:
            return None
        return Verified(**status_info)

    @validator("false_negative", always=True)
    def set_false_negative(cls, v, values: dict) -> Optional[FalseNegative]:
        status_info: Optional[dict] = values["status_data_"].get("False Negative")
        if not status_info:
            return None
        return FalseNegative(**status_info)

    @validator("other", always=True)
    def set_other(cls, v, values: dict) -> Optional[Other]:
        status_info: Optional[dict] = values["status_data_"].get("Other")
        if not status_info:
            return None
        return Other(**status_info)

    @validator("failed", always=True)
    def set_failed(cls, v, values: dict) -> Optional[Failed]:
        status_info: Optional[dict] = values["status_data_"].get("Failed")
        if not status_info:
            return None
        return Failed(**status_info)


class FetalFractionControlAbNormal(BaseModel):
    """Model for sex chromosome abnormalities.

    This model is only for samples that have ben classified as abnormal sex chromosomes in some sense.

    Samples are grouped by sex chromosome abnormality. Samples within a abnormality group are then grouped by
    abnormality classification, defined by AbNormalityClasses.

    Eg; Each sample in XO have been classified as not normal for sex chromosome abnormality X0.
    The samples within X0 are then grouped by Verified X0, False Positive X0, etc"""

    X0: AbNormalityClasses
    XXX: AbNormalityClasses
    XXY: AbNormalityClasses
    XYY: AbNormalityClasses

    class Config:
        validate_assignment = True



from datetime import datetime
from typing import Literal

from pydantic import BaseModel, EmailStr


class User(BaseModel):
    username: str
    email: EmailStr
    added: datetime
    role: Literal["R", "RW", "inactive", "admin"]
    hashed_password: str



import logging
from typing import Dict, List, Optional

from pymongo.results import InsertManyResult, InsertOneResult

from statina.adapter import StatinaAdapter
from statina.exeptions import InsertError
from statina.models.database import Batch, DataBaseSample, User
from statina.models.server.load import BatchRequestBody
from statina.parse.batch import parse_segmental_calls

LOG = logging.getLogger(__name__)


def insert_batch(adapter: StatinaAdapter, batch: Batch, batch_files: BatchRequestBody) -> str:
    """Load a batch into the database"""

    batch_dict = batch.dict(exclude_none=True)
    batch_dict["segmental_calls"] = batch_files.segmental_calls
    batch_dict["multiqc_report"] = batch_files.multiqc_report
    batch_dict["result_file"] = batch_files.result_file
    try:
        result: InsertOneResult = adapter.batch_collection.insert_one(batch_dict)
        LOG.info("Added document %s.", batch_dict["batch_id"])
    except:
        raise InsertError(message=f"Batch {batch.batch_id} already in database.")
    return result.inserted_id


def insert_samples(
    adapter: StatinaAdapter, samples: List[DataBaseSample], segmental_calls: Optional[str]
) -> List[str]:
    """Function to load data from fluffy result file."""

    segmental_calls: Dict[str, str] = parse_segmental_calls(segmental_calls_path=segmental_calls)
    sample_dicts = []
    for sample in samples:
        sample_dict: dict = sample.dict(exclude_none=True)
        sample_dict["segmental_calls"] = segmental_calls.get(sample.sample_id)
        sample_dicts.append(sample_dict)
    try:
        result: InsertManyResult = adapter.sample_collection.insert_many(sample_dicts)
        LOG.info("Added sample documents.")
    except:
        raise InsertError(f"Sample keys already in database.")
    return result.inserted_ids


def insert_user(adapter: StatinaAdapter, user: User) -> str:
    """Function to load a new user to the database."""

    existing_user: Optional[dict] = adapter.user_collection.find(
        {"$or": [{"username": user.username}, {"email": user.email}]}
    )
    if list(existing_user):
        raise InsertError(
            f"User with username {user.username} or email {user.email} already exist in database."
        )
    try:
        result: InsertOneResult = adapter.user_collection.insert_one(user.dict())
        LOG.info("Added user document %s.", user.email)
    except:
        raise InsertError(f"User {user.email} already in database.")
    return result.inserted_id



from typing import Dict, List

from fastapi import APIRouter, Depends, Request

import statina.crud.find.plots.fetal_fraction_plot_data as get_fetal_fraction
from statina.adapter import StatinaAdapter
from statina.API.external.api.deps import get_current_user
from statina.API.external.constants import TRISOMI_TRESHOLDS, COLORS
from statina.config import get_nipt_adapter, templates
from statina.crud.find import find
from statina.crud.find.plots.coverage_plot_data import (
    get_box_data_for_coverage_plot,
    get_scatter_data_for_coverage_plot,
)
from statina.crud.find.plots.zscore_plot_data import (
    get_tris_control_abnormal,
    get_tris_control_normal,
    get_tris_samples,
)
from statina.models.database import Batch, DataBaseSample, User
from statina.models.server.plots.coverage import CoveragePlotSampleData
from statina.models.server.plots.fetal_fraction import (
    FetalFractionControlAbNormal,
    FetalFractionSamples,
)
from statina.models.server.plots.fetal_fraction_sex import SexChromosomeThresholds
from statina.models.server.sample import Sample

router = APIRouter()


@router.post("/")
def batches(
    request: Request,
    adapter: StatinaAdapter = Depends(get_nipt_adapter),
    user: User = Depends(get_current_user),
):
    """List of all batches"""

    all_batches: List[Batch] = find.batches(adapter=adapter)
    return templates.TemplateResponse(
        "batches.html",
        context={
            "request": request,
            "batches": all_batches,
            "current_user": user,
            "page_id": "all_batches",
        },
    )


@router.get("/")
def batches(
    request: Request,
    adapter: StatinaAdapter = Depends(get_nipt_adapter),
    user: User = Depends(get_current_user),
):
    """List of all batches"""
    all_batches: List[Batch] = find.batches(adapter=adapter)
    return templates.TemplateResponse(
        "batches.html",
        context={
            "request": request,
            "batches": all_batches,
            "current_user": user,
            "page_id": "all_batches",
        },
    )


@router.post("/{batch_id}/")
def batch(
    request: Request,
    batch_id: str,
    adapter: StatinaAdapter = Depends(get_nipt_adapter),
    user: User = Depends(get_current_user),
):
    """Batch view with table of all samples in the batch."""

    samples: List[DataBaseSample] = find.batch_samples(batch_id=batch_id, adapter=adapter)
    return templates.TemplateResponse(
        "batch/tabs/table.html",
        context={
            "request": request,
            "batch": find.batch(batch_id=batch_id, adapter=adapter),
            "sample_info": [Sample(**sample.dict()) for sample in samples],
            "page_id": "batches",
            "current_user": user,
        },
    )


@router.get("/{batch_id}/")
def batch(
    request: Request,
    batch_id: str,
    adapter: StatinaAdapter = Depends(get_nipt_adapter),
    user: User = Depends(get_current_user),
):
    """Batch view with table of all samples in the batch."""

    samples: List[DataBaseSample] = find.batch_samples(batch_id=batch_id, adapter=adapter)
    return templates.TemplateResponse(
        "batch/tabs/table.html",
        context={
            "request": request,
            "batch": find.batch(batch_id=batch_id, adapter=adapter),
            "sample_info": [Sample(**sample.dict()) for sample in samples],
            "page_id": "batches",
            "current_user": user,
        },
    )


@router.get("/{batch_id}/{ncv}")
def Zscore(
    request: Request,
    batch_id: str,
    ncv: str,
    adapter: StatinaAdapter = Depends(get_nipt_adapter),
    user: User = Depends(get_current_user),
):
    """Batch view with with Zscore plot"""

    batch: Batch = find.batch(batch_id=batch_id, adapter=adapter)

    return templates.TemplateResponse(
        "batch/tabs/Zscore.html",
        context=dict(
            request=request,
            tris_thresholds=TRISOMI_TRESHOLDS,
            batch=batch.dict(),
            chromosomes=[ncv],
            ncv_chrom_data={ncv: get_tris_samples(adapter=adapter, chr=ncv, batch_id=batch_id)},
            normal_data={ncv: get_tris_control_normal(adapter, ncv)},
            abnormal_data={ncv: get_tris_control_abnormal(adapter, ncv, 0)},
            page_id=f"batches_NCV{ncv}",
            current_user=user,
        ),
    )


@router.get("/batches/{batch_id}/fetal_fraction_XY")
def fetal_fraction_XY(
    request: Request,
    batch_id: str,
    adapter: StatinaAdapter = Depends(get_nipt_adapter),
    user: User = Depends(get_current_user),
):
    """Batch view with fetal fraction (X against Y) plot"""

    batch: Batch = find.batch(batch_id=batch_id, adapter=adapter)

    cases = get_fetal_fraction.samples(adapter=adapter, batch_id=batch_id)
    control: FetalFractionSamples = get_fetal_fraction.samples(
        batch_id=batch_id, adapter=adapter, control_samples=True
    )
    abnormal: FetalFractionControlAbNormal = get_fetal_fraction.control_abnormal(adapter)
    abnormal_dict = abnormal.dict(
        exclude_none=True,
        exclude={
            "X0": {"status_data_"},
            "XXX": {"status_data_"},
            "XXY": {"status_data_"},
            "XYY": {"status_data_"},
        },
    )

    x_max = max(control.FFX + cases.FFX) + 1
    x_min = min(control.FFX + cases.FFX) - 1

    sex_thresholds = SexChromosomeThresholds(x_min=x_min, x_max=x_max)
    return templates.TemplateResponse(
        "batch/tabs/FF_XY.html",
        context=dict(
            sex_thresholds={
                "XY_fetal_fraction_y": sex_thresholds.XY_fetal_fraction_y(),
                "XX_lower": sex_thresholds.XX_lower(),
                "XX_upper": sex_thresholds.XX_upper(),
                "XY_upper": sex_thresholds.XY_upper(),
                "XY_lower": sex_thresholds.XY_lower(),
                "XXY": sex_thresholds.XXY(),
            },
            request=request,
            current_user=user,
            colors=COLORS,
            control=control,
            abnormal=abnormal_dict,
            cases=cases,
            max_x=x_max,
            min_x=x_min,
            batch=batch.dict(),
            page_id="batches_FF_XY",
        ),
    )


@router.get("/batches/{batch_id}/fetal_fraction")
def fetal_fraction(
    request: Request,
    batch_id: str,
    adapter: StatinaAdapter = Depends(get_nipt_adapter),
    user: User = Depends(get_current_user),
):
    """Batch view with fetal fraction plot"""
    batch: Batch = find.batch(batch_id=batch_id, adapter=adapter)
    return templates.TemplateResponse(
        "batch/tabs/FF.html",
        context=dict(
            request=request,
            current_user=user,
            colors=COLORS,
            control=get_fetal_fraction.samples(
                adapter=adapter, batch_id=batch_id, control_samples=True
            ),
            cases=get_fetal_fraction.samples(adapter=adapter, batch_id=batch_id),
            batch=batch.dict(),
            page_id="batches_FF",
        ),
    )


@router.get("/batches/{batch_id}/coverage")
def coverage(
    request: Request,
    batch_id: str,
    adapter: StatinaAdapter = Depends(get_nipt_adapter),
    user: User = Depends(get_current_user),
):
    """Batch view with coverage plot"""
    batch: Batch = find.batch(batch_id=batch_id, adapter=adapter)
    db_samples: List[DataBaseSample] = find.batch_samples(batch_id=batch_id, adapter=adapter)
    samples: List[Sample] = [Sample(**db_sample.dict()) for db_sample in db_samples]

    scatter_data: Dict[str, CoveragePlotSampleData] = get_scatter_data_for_coverage_plot(samples)
    box_data: Dict[int, List[float]] = get_box_data_for_coverage_plot(samples)
    return templates.TemplateResponse(
        "batch/tabs/coverage.html",
        context=dict(
            request=request,
            current_user=user,
            batch=batch.dict(),
            x_axis=list(range(1, 23)),
            scatter_data=scatter_data,
            box_data=box_data,
            page_id="batches_cov",
        ),
    )



from fastapi import FastAPI
from fastapi.responses import RedirectResponse, Response
from fastapi import Request

from statina.API.external.api.api_v1.endpoints import (
    batches,
    download,
    index,
    login,
    user,
    sample,
    statistics,
    update,
)
from statina.API.internal.api.api_v1.endpoints import insert
from statina.exeptions import CredentialsError

external_app = FastAPI(
    servers=[{"url": "https://nipttol-stage.scilifelab.se", "description": "Staging environment"}],
    root_path_in_servers=False,
)


@external_app.exception_handler(CredentialsError)
async def exception_handler(request: Request, exc: CredentialsError) -> Response:
    return RedirectResponse(url="/")


external_app.include_router(login.router, prefix="/login", tags=["login"])
external_app.include_router(batches.router, prefix="/batches", tags=["batches"])
external_app.include_router(index.router, tags=["index"])
external_app.include_router(sample.router, tags=["sample"])
external_app.include_router(update.router, tags=["update"])
external_app.include_router(download.router, tags=["download"])
external_app.include_router(statistics.router, tags=["statistics"])

internal_app = FastAPI()
internal_app.include_router(insert.router, prefix="/insert", tags=["insert"])



from typing import Iterable

import logging

from statina.adapter import StatinaAdapter
from statina.crud.find.find import batch_samples


LOG = logging.getLogger(__name__)


def delete_batch_document(adapter: StatinaAdapter, batch_id: str):
    adapter.batch_collection.delete_one({"batch_id": batch_id})
    LOG.info(f"Deleting batch {batch_id}")


def delete_sample_document(adapter: StatinaAdapter, sample_id: str):
    adapter.sample_collection.delete_one({"sample_id": sample_id})
    LOG.info(f"Deleting sample {sample_id}")


def delete_batch(adapter: StatinaAdapter, batch_id: str):
    samples = batch_samples(adapter=adapter, batch_id=batch_id)
    for sample in samples:
        delete_sample_document(adapter=adapter, sample_id=sample.sample_id)
    delete_batch_document(adapter=adapter, batch_id=batch_id)


def delete_batches(adapter: StatinaAdapter, batches: Iterable[str]):
    for batch_id in batches:
        delete_batch(adapter=adapter, batch_id=batch_id)



from typing import Literal, Optional

from pydantic import BaseModel, validator

from statina.API.external.constants import (
    SEX_CHROM_ABNORM,
    TRIS_CHROM_ABNORM,
    TRISOMI_TRESHOLDS,
    FF_TRESHOLDS,
)
from statina.models.database import DataBaseSample
from statina.models.server.plots.fetal_fraction_sex import x_get_y


class SampleWarning(BaseModel):
    FF_Formatted: Literal["danger", "default", "warning"]
    Zscore_13: Literal["danger", "default", "warning"]
    Zscore_18: Literal["danger", "default", "warning"]
    Zscore_21: Literal["danger", "default", "warning"]
    FFY: Literal["danger", "default", "warning"]
    X0: Literal["danger", "default", "warning"]
    XXX: Literal["danger", "default", "warning"]
    other: Literal["danger", "default", "warning"]
    XXY: Literal["danger", "default", "warning"]
    XYY: Literal["danger", "default", "warning"]


class Sample(DataBaseSample):
    warnings: Optional[SampleWarning]
    text_warning: Optional[str]
    status: Optional[str]
    sex: Optional[Literal["XX", "XY"]]

    @validator("Zscore_13")
    def round_zscore_13(cls, v):
        return round(v, 2)

    @validator("Zscore_18")
    def round_zscore_18(cls, v):
        return round(v, 2)

    @validator("Zscore_21")
    def round_zscore_21(cls, v):
        return round(v, 2)

    @validator("Zscore_X")
    def round_zscore_x(cls, v):
        return round(v, 2)

    @validator("FF_Formatted")
    def round_ff(cls, v):
        return round(v, 2)

    @validator("FFX")
    def round_ffx(cls, v):
        return round(v, 2)

    @validator("FFY")
    def round_ffy(cls, v):
        return round(v, 2)

    @validator("status", always=True)
    def set_status(cls, v, values: dict) -> str:

        status_list = []
        for key in TRIS_CHROM_ABNORM + SEX_CHROM_ABNORM:
            status = values.get(f"status_{key}")
            if status and status != "Normal":
                status_list.append(" ".join([status, key]))
        return ", ".join(status_list)

    @validator("warnings", always=True)
    def set_warnings(cls, v, values: dict) -> SampleWarning:
        """Get warnings for a sample and return a SampleWarning"""

        sample_warnings = {}
        fetal_fraction_pf = values.get("FF_Formatted")
        fetal_fraction_y = values.get("FFY")
        fetal_fraction_x = values.get("FFX")
        sample_warnings["FF_Formatted"]: str = cls.get_ff_preface_warning(
            fetal_fraction_pf=fetal_fraction_pf, fetal_fraction_y=fetal_fraction_y
        )
        sample_warnings["other"]: str = cls.get_other_warning(
            fetal_fraction_y=fetal_fraction_y, fetal_fraction_x=fetal_fraction_x
        )
        sample_warnings["FFY"]: str = cls.get_ff_y_warning(fetal_fraction_y=fetal_fraction_y)
        sample_warnings["X0"]: str = cls.get_x0_warning(
            fetal_fraction_y=fetal_fraction_y, fetal_fraction_x=fetal_fraction_x
        )
        sample_warnings["XXX"]: str = cls.get_XXX_warning(
            fetal_fraction_y=fetal_fraction_y, fetal_fraction_x=fetal_fraction_x
        )
        sample_warnings["XYY"]: str = cls.get_XYY_warning(
            fetal_fraction_y=fetal_fraction_y, fetal_fraction_x=fetal_fraction_x
        )
        sample_warnings["XXY"]: str = cls.get_XXY_warning(
            fetal_fraction_y=fetal_fraction_y, fetal_fraction_x=fetal_fraction_x
        )

        for key in ["Zscore_13", "Zscore_18", "Zscore_21"]:
            z_score = values.get(key)
            sample_warnings[key]: str = cls.get_tris_warning(
                z_score=z_score, fetal_fraction=fetal_fraction_pf
            )
        return SampleWarning(**sample_warnings)

    @validator("text_warning", always=True)
    def set_text_warning(cls, v, values: dict) -> str:
        """Joining the warnings for a sample to a text string"""

        text_warnings = [
            abn for abn, warning in values["warnings"].dict().items() if warning == "danger"
        ]
        return ", ".join(text_warnings)

    @validator("sex", always=True)
    def set_sex(cls, v, values: dict) -> str:
        """Set sex based on fetal fraction Y thresholds"""

        fetal_fraction_y = values.get("FFY")
        return "XY" if fetal_fraction_y >= FF_TRESHOLDS["fetal_fraction_y_min"] else "XX"

    @classmethod
    def get_tris_warning(cls, z_score: float, fetal_fraction: float) -> str:
        """Get automated trisomy warning, based on preset Zscore thresholds"""

        hard_max = TRISOMI_TRESHOLDS["hard_max"]["Zscore"]
        soft_max = TRISOMI_TRESHOLDS["soft_max"]["Zscore"]
        soft_min = TRISOMI_TRESHOLDS["soft_min"]["Zscore"]
        hard_min = TRISOMI_TRESHOLDS["hard_min"]["Zscore"]
        preface = FF_TRESHOLDS["fetal_fraction_preface"]

        if fetal_fraction is None or z_score is None:
            return "default"
        elif fetal_fraction < preface and soft_max < z_score < hard_max:
            return "warning"
        elif hard_min < z_score <= soft_min:
            return "warning"
        elif z_score >= hard_max or z_score <= hard_min:
            return "danger"
        else:
            return "default"

    @classmethod
    def get_ff_y_warning(cls, fetal_fraction_y: float) -> str:
        """Get fetal fraction warning based on preset threshold"""

        y_min = FF_TRESHOLDS["fetal_fraction_y_min"]
        y_max = FF_TRESHOLDS["fetal_fraction_y_max"]

        if not isinstance(fetal_fraction_y, (float, int)):
            return "default"

        if y_min <= fetal_fraction_y < y_max:
            return "danger"

        return "default"

    @classmethod
    def get_x0_warning(cls, fetal_fraction_y: float, fetal_fraction_x: float) -> str:
        """Get fetal fraction warning based on preset threshold"""

        x_treshold = FF_TRESHOLDS["fetal_fraction_X0"]
        y_treshold = FF_TRESHOLDS["fetal_fraction_y_min"]

        if not (
            isinstance(fetal_fraction_y, (float, int))
            and isinstance(fetal_fraction_x, (float, int))
        ):
            return "default"

        if fetal_fraction_y < y_treshold and fetal_fraction_x > x_treshold:
            return "danger"

        return "default"

    @classmethod
    def get_XXX_warning(cls, fetal_fraction_y: float, fetal_fraction_x: float) -> str:
        """Get fetal fraction warning based on preset threshold"""

        x_treshold = FF_TRESHOLDS["fetal_fraction_XXX"]
        y_treshold = FF_TRESHOLDS["fetal_fraction_y_min"]

        if not (
            isinstance(fetal_fraction_y, (float, int))
            and isinstance(fetal_fraction_x, (float, int))
        ):
            return "default"

        if fetal_fraction_y < y_treshold and fetal_fraction_x < x_treshold:
            return "danger"

        return "default"

    @classmethod
    def get_ff_preface_warning(cls, fetal_fraction_pf: float, fetal_fraction_y: float) -> str:
        """Get fetal fraction preface warning based on preset threshold"""

        if not (
            isinstance(fetal_fraction_y, (float, int))
            and isinstance(fetal_fraction_pf, (float, int))
        ):
            return "default"
        y_min = FF_TRESHOLDS["fetal_fraction_y_min"]
        pf_min = FF_TRESHOLDS["fetal_fraction_preface"]
        if fetal_fraction_pf < pf_min and fetal_fraction_y < y_min:
            return "danger"
        return "default"

    @classmethod
    def get_other_warning(cls, fetal_fraction_y: float, fetal_fraction_x: float) -> str:
        """Get fetal fraction warning based on preset threshold"""

        k_lower: float = FF_TRESHOLDS["k_lower"]
        m_lower: float = FF_TRESHOLDS["m_lower"]
        y_treshold: float = FF_TRESHOLDS["fetal_fraction_y_min"]

        if not (
            isinstance(fetal_fraction_y, (float, int))
            and isinstance(fetal_fraction_x, (float, int))
        ):
            return "default"

        if y_treshold < fetal_fraction_y <= x_get_y(x=fetal_fraction_x, k=k_lower, m=m_lower):
            return "danger"

        return "default"

    @classmethod
    def get_XXY_warning(cls, fetal_fraction_y: float, fetal_fraction_x: float) -> str:
        """Get fetal fraction warning based on preset threshold"""

        k_upper: float = FF_TRESHOLDS["k_upper"]
        m_upper: float = FF_TRESHOLDS["m_upper"]
        x_treshold: float = FF_TRESHOLDS["fetal_fraction_X0"]

        if not (
            isinstance(fetal_fraction_y, (float, int))
            and isinstance(fetal_fraction_x, (float, int))
        ):
            return "default"

        if fetal_fraction_y > x_get_y(x=fetal_fraction_x, k=k_upper, m=m_upper) and (
            fetal_fraction_x <= x_treshold
        ):
            return "danger"
        return "default"

    @classmethod
    def get_XYY_warning(cls, fetal_fraction_y: float, fetal_fraction_x: float) -> str:
        """Get fetal fraction warning based on preset threshold"""

        k_upper: float = FF_TRESHOLDS["k_upper"]
        m_upper: float = FF_TRESHOLDS["m_upper"]
        x_treshold: float = FF_TRESHOLDS["fetal_fraction_X0"]

        if not (
            isinstance(fetal_fraction_y, (float, int))
            and isinstance(fetal_fraction_x, (float, int))
        ):
            return "default"

        if fetal_fraction_y > x_get_y(x=fetal_fraction_x, k=k_upper, m=m_upper) and (
            fetal_fraction_x > x_treshold
        ):
            return "danger"
        return "default"



import logging
from datetime import datetime
from typing import Iterable

from fastapi import APIRouter, Depends, Request
from fastapi.responses import RedirectResponse
from starlette.datastructures import FormData

from statina.adapter import StatinaAdapter
from statina.API.external.api.deps import get_current_user
from statina.API.external.constants import CHROM_ABNORM
from statina.config import get_nipt_adapter
from statina.crud import update
from statina.crud.delete import delete_batches
from statina.crud.find import find
from statina.models.database import DataBaseSample, User, Batch

router = APIRouter()

LOG = logging.getLogger(__name__)


@router.post("/update_user")
async def update_user(
    request: Request,
    adapter: StatinaAdapter = Depends(get_nipt_adapter),
    user: User = Depends(get_current_user),
):
    if user.role != "admin":
        return RedirectResponse(request.headers.get("referer"))
    form = await request.form()
    update_user: User = find.user(email=form["user_email"], adapter=adapter)
    update_user.role = form["role"]
    update.update_user(adapter=adapter, user=update_user)
    return RedirectResponse(request.headers.get("referer"))


@router.post("/delete_batch")
async def delete_batch(
    request: Request,
    adapter: StatinaAdapter = Depends(get_nipt_adapter),
    user: User = Depends(get_current_user),
):
    if user.role != "admin":
        return RedirectResponse(request.headers.get("referer"))
    form = await request.form()
    batches: Iterable[str] = form.getlist("delete")
    delete_batches(adapter=adapter, batches=batches)
    return RedirectResponse(request.headers.get("referer"))


@router.post("/set_sample_status")
async def set_sample_status(
    request: Request,
    adapter: StatinaAdapter = Depends(get_nipt_adapter),
    user: User = Depends(get_current_user),
):
    """Update the manualy interpreted chromosome abnormality status for a sample."""

    form = await request.form()

    if user.role not in ["RW", "admin"]:
        return RedirectResponse(request.headers.get("referer"))

    sample_id: str = form["sample_id"]
    sample: dict = find.sample(sample_id=sample_id, adapter=adapter).dict()

    for abnormality in CHROM_ABNORM:
        new_abnormality_status: str = form[abnormality]
        abnormality_key: str = f"status_{abnormality}"
        if sample.get(abnormality_key) == new_abnormality_status:
            continue

        LOG.debug(
            "Updating %s to %s for sample %s",
            abnormality_key,
            new_abnormality_status,
            sample_id,
        )
        time_stamp: str = datetime.now().strftime("%Y/%m/%d %H:%M:%S")
        sample[abnormality_key] = new_abnormality_status
        sample[f"status_change_{abnormality}"] = f"{user.username} {time_stamp}"

    update.sample(adapter=adapter, sample=DataBaseSample(**sample))
    return RedirectResponse(request.headers.get("referer"))


@router.post("/batch_comment")
async def batch_comment(
    request: Request,
    adapter: StatinaAdapter = Depends(get_nipt_adapter),
    user: User = Depends(get_current_user),
):
    """Update batch comment"""

    form = await request.form()

    if user.role not in ["RW", "admin"]:
        return RedirectResponse(request.headers.get("referer"))
    batch_id: str = form["batch_id"]
    batch: Batch = find.batch(batch_id=batch_id, adapter=adapter)
    comment: str = form.get("comment")
    if comment != batch.comment:
        batch.comment = comment
        update.update_batch(adapter=adapter, batch=batch)

    return RedirectResponse(request.headers.get("referer"))


@router.post("/sample_comment")
async def sample_comment(
    request: Request,
    adapter: StatinaAdapter = Depends(get_nipt_adapter),
    user: User = Depends(get_current_user),
):
    """Update sample comment"""

    form = await request.form()

    if user.role not in ["RW", "admin"]:
        return RedirectResponse(request.headers.get("referer"))

    sample_id: str = form["sample_id"]
    sample: DataBaseSample = find.sample(sample_id=sample_id, adapter=adapter)
    comment: str = form.get("comment")
    if comment != sample.comment:
        sample.comment = comment
        update.sample(adapter=adapter, sample=sample)

    return RedirectResponse(request.headers.get("referer"))


@router.post("/include_samples")
async def include_samples(
    request: Request,
    adapter: StatinaAdapter = Depends(get_nipt_adapter),
    user: User = Depends(get_current_user),
):
    """Update include status and comment for samples in batch"""

    form = await request.form()

    if user.role not in ["RW", "admin"]:
        return RedirectResponse(request.headers.get("referer"))

    button_id = form.get("button_id")
    samples: Iterable[str] = form.getlist("samples")
    if button_id == "include all samples":
        include_all_samples(samples=samples, adapter=adapter, user=user)
    elif button_id == "Save":
        save_samples(samples=samples, form=form, adapter=adapter, user=user)

    return RedirectResponse(request.headers.get("referer"))


def save_samples(samples: Iterable[str], form: FormData, adapter: StatinaAdapter, user: User):
    """Function to update sample.comment and sample.include."""

    time_stamp: str = datetime.now().strftime("%Y/%m/%d %H:%M:%S")
    for sample_id in samples:
        sample: DataBaseSample = find.sample(sample_id=sample_id, adapter=adapter)
        comment: str = form.get(f"comment_{sample_id}")
        include: bool = form.get(f"include_{sample_id}")
        if comment != sample.comment:
            sample.comment = comment
        if include and not sample.include:
            sample.include = True
            sample.change_include_date = f"{user.username} {time_stamp}"
        elif not include and sample.include:
            sample.include = False
        update.sample(adapter=adapter, sample=sample)


def include_all_samples(samples: Iterable[str], adapter: StatinaAdapter, user: User):
    """Function to set sample.include=True for all samples."""

    time_stamp: str = datetime.now().strftime("%Y/%m/%d %H:%M:%S")
    for sample_id in samples:
        sample: DataBaseSample = find.sample(sample_id=sample_id, adapter=adapter)
        if sample.include:
            continue
        sample.include = True
        sample.change_include_date = f"{user.username} {time_stamp}"
        update.sample(adapter=adapter, sample=sample)



from statina.adapter.plugin import StatinaAdapter
from statina.crud.insert import insert_batch, insert_user
from statina.parse.batch import get_batch, get_samples


def test_load_user(database, valid_load_user):
    # GIVEN a user with valid requiered fields and a statina database adapter
    nipt_adapter = StatinaAdapter(database.client, db_name="test")

    # WHEN running insert_user
    insert_user(nipt_adapter, valid_load_user)

    # The user should be added to the database
    assert nipt_adapter.user_collection.estimated_document_count() == 1


def test_batch_valid_files(database, valid_csv, valid_load_batch):
    # GIVEN the following request data:
    #   a valid csv file with three samples
    #   segmental_calls and multiqc_report files with random content, but that do exist.
    nipt_adapter = StatinaAdapter(database.client, db_name="test")
    batch = get_batch(valid_csv)

    # WHEN running insert_batch
    insert_batch(nipt_adapter, batch, valid_load_batch)

    # THEN the batch should be added to the batch collection
    assert nipt_adapter.batch_collection.estimated_document_count() == 1


"""def test_samples_valid_files(database, valid_csv, segmental_calls, multiqc_report):
    # GIVEN the following request data:
    #   a valid csv file with three samples
    #   segmental_calls and multiqc_report files with random content, but that do exist.

    data = dict(multiqc_report=multiqc_report, segmental_calls=segmental_calls, result_file=valid_csv)

    # WHEN running the request with the data
    response = fast_app_client.post('/batch', data=data)

    # THEN assert the samples should be added to the sample collection
    # and the batch should be added to the batch collection
    assert mock_app.adapter.sample_collection.estimated_document_count() == 3
    assert mock_app.adapter.batch_collection.estimated_document_count() == 1
    resp_data = json.loads(response.data)
    assert resp_data["message"] == "Data loaded into database"
    assert response.status_code == 200"""



import pytest
from fastapi import status
from pydantic import ValidationError

"""
def test_user_mocked(mocker, fast_app_client, valid_load_user):
    # GIVEN a insert_user function that never fails
    mocker.patch('statina.server.load.api.api_v1.endpoints.load.insert_user')

    # WHEN running the user request with the data
    response = fast_app_client.post('/api/v1/insert/user', json=valid_load_user.dict())

    assert response.status_code == status.HTTP_200_OK


def test_user_load_fails_mocked(mocker, mock_fast_client):
    # GIVEN the following request data:
    data = dict(email='maya.papaya@something.se', name="Maya Papaya", role="RW")

    with patch('statina.server.load.api.api_v1.endpoints.load.insert_user', side_effect=Exception('mocked error')):
        # WHEN running the user request with the data
        response = mock_fast_client.post('/api/v1/insert/user', json=data)
        assert response.status_code == status.HTTP_422_UNPROCESSABLE_ENTITY"""


# def test_user(fast_app_client, valid_load_user):
#     # GIVEN valid user load data
#
#     # WHEN running the user request with the data
#     response = fast_app_client.post("/api/v1/insert/user", json=valid_load_user.dict())
#
#     # THEN
#     assert "inserted to the database" in response.text
#     assert response.status_code == status.HTTP_200_OK


# def test_user_wrong_input(fast_app_client, valid_load_user):
#     # GIVEN user data with missing email field
#     data = valid_load_user.dict()
#     data.pop("email")
#
#     # WHEN running the user request with the data
#     response = fast_app_client.post("/api/v1/insert/user", json=data)
#
#     # THEN
#     assert response.status_code == status.HTTP_422_UNPROCESSABLE_ENTITY


# def test_batch_valid_files(fast_app_client, valid_load_batch):
#     # GIVEN the valid load batch data
#
#     # WHEN running the request with the data
#     response = fast_app_client.post("/api/v1/insert/batch", json=valid_load_batch.dict())
#
#     # THEN
#     assert "inserted to the database" in response.text
#     assert response.status_code == status.HTTP_200_OK


# def test_batch_wrong_segmental_calls_path(fast_app_client, valid_load_batch):
#     # GIVEN the following request data:
#     #   a valid csv file with three samples
#     #   but no segmental_calls and multiqc_report
#
#     data = valid_load_batch.dict()
#     data["segmental_calls"] = "no_valid_path"
#
#     # WHEN running the request with the data
#     response = fast_app_client.post("/api/v1/insert/batch", json=data)
#
#     # THEN assert batch is being added without error
#     assert "inserted to the database" in response.text
#     assert response.status_code == status.HTTP_200_OK


# def test_batch_wrong_result_file_path(fast_app_client, valid_load_batch):
#     # GIVEN the following request data:
#
#     data = valid_load_batch.dict()
#     data["result_file"] = "no_valid_path"
#
#     # WHEN running the request with the data
#     response = fast_app_client.post("/api/v1/insert/batch", json=data)
#
#     # THEN assert
#     assert response.status_code == status.HTTP_422_UNPROCESSABLE_ENTITY
#     assert "Results file missing" in response.text


# def test_batch_no_data(fast_app_client):
#     # GIVEN no data
#
#     # WHEN running the request with empty data
#     response = fast_app_client.post("/api/v1/insert/batch", json=dict())
#
#     # THEN assert BadRequestKeyError: 400 Bad Request
#     assert response.status_code == status.HTTP_422_UNPROCESSABLE_ENTITY


# def test_batch_invalid_file(fast_app_client, valid_load_batch, csv_with_missing_sample_id):
#     # GIVEN batch load data with a csv result file with missing sample ids:
#     data = valid_load_batch.dict()
#     data["result_file"] = csv_with_missing_sample_id
#
#     # WHEN running the request with the data
#     # THEN assert ValidationError and "Required fields missing" in response message
#     with pytest.raises(ValidationError):
#         response = fast_app_client.post("/api/v1/insert/batch", json=data)
#         assert "Could not load data from result file. Required fields missing." in response.text



from fastapi import APIRouter, Depends, Request

from statina.adapter.plugin import StatinaAdapter
from statina.API.external.api.deps import get_current_user
from statina.config import get_nipt_adapter, templates
from statina.crud.find.plots.statistics_plot_data import (
    get_last_batches,
    get_statistics_for_box_plot,
    get_statistics_for_scatter_plot,
)
from statina.models.database import User

router = APIRouter()


@router.get("/statistics")
def statistics(
    request: Request,
    adapter: StatinaAdapter = Depends(get_nipt_adapter),
    user: User = Depends(get_current_user),
):
    """Statistics view."""

    nr_batches = 20
    scatter_plots = ["Stdev_13", "Stdev_18", "Stdev_21"]
    box_plots = [
        "Chr13_Ratio",
        "Chr18_Ratio",
        "Chr21_Ratio",
        "FF_Formatted",
        "DuplicationRate",
        "MappedReads",
        "GC_Dropout",
        "AT_Dropout",
        "Bin2BinVariance",
        "Library_nM",
    ]

    batches = get_last_batches(adapter=adapter, nr_of_batches=nr_batches)
    batch_ids = [batch.get("batch_id") for batch in batches]
    box_stat = get_statistics_for_box_plot(adapter=adapter, batches=batch_ids, fields=box_plots)
    scatter_stat = get_statistics_for_scatter_plot(batches=batches, fields=scatter_plots)
    return templates.TemplateResponse(
        "statistics.html",
        context=dict(
            request=request,
            current_user=user,
            ticks=list(range(0, nr_batches)),
            nr_batches=nr_batches,
            batch_ids=batch_ids,
            box_stat=box_stat,
            box_plots=box_plots,
            scatter_stat=scatter_stat,
            scatter_plots=scatter_plots,
            page_id="statistics",
        ),
    )



SAMPLE_KEYS = [
    "SampleProject",
    "SampleType",
    "Description",
    "Flowcell",
    "Index1",
    "Index2",
    "Library_nM",
    "QCFlag",
    "Zscore_13",
    "Zscore_18",
    "Zscore_21",
    "Zscore_X",
    "MappedReads",
    "GC_Dropout",
    "AT_Dropout",
    "Chr1_Ratio",
    "Chr2_Ratio",
    "Chr3_Ratio",
    "Chr4_Ratio",
    "Chr5_Ratio",
    "Chr6_Ratio",
    "Chr7_Ratio",
    "Chr8_Ratio",
    "Chr9_Ratio",
    "Chr10_Ratio",
    "Chr11_Ratio",
    "Chr12_Ratio",
    "Chr13_Ratio",
    "Chr18_Ratio",
    "Chr21_Ratio",
    "ChrX_Ratio",
    "ChrY_Ratio",
    "Chr14_Ratio",
    "Chr15_Ratio",
    "Chr16_Ratio",
    "Chr17_Ratio",
    "Chr19_Ratio",
    "Chr20_Ratio",
    "Chr22_Ratio",
    "Chr1",
    "Chr2",
    "Chr3",
    "Chr4",
    "Chr5",
    "Chr6",
    "Chr7",
    "Chr8",
    "Chr9",
    "Chr10",
    "Chr11",
    "Chr12",
    "Chr13",
    "Chr14",
    "Chr15",
    "Chr16",
    "Chr17",
    "Chr18",
    "Chr19",
    "Chr20",
    "Chr21",
    "Chr22",
    "ChrX",
    "ChrY",
    "FF_Formatted",
    "FFY",
    "FFX",
    "DuplicationRate",
    "Bin2BinVariance",
    "UnfilteredCNVcalls",
    "CNVSegment",
]

BATCH_KEYS = [
    "Flowcell",
    "SequencingDate",
    "Median_13",
    "Median_18",
    "Median_21",
    "Median_X",
    "Median_Y",
    "Stdev_13",
    "Stdev_18",
    "Stdev_21",
    "Stdev_X",
    "Stdev_Y",
]



import logging
from statina.adapter import StatinaAdapter
from statina.models.database import DataBaseSample, User, Batch

LOG = logging.getLogger(__name__)


def sample(adapter: StatinaAdapter, sample: DataBaseSample) -> None:
    """Update a sample object in the database"""

    sample_dict: dict = sample.dict(exclude_none=True)
    sample_id = sample.sample_id
    LOG.info("Updating sample %s", sample_id)
    adapter.sample_collection.update_one({"sample_id": sample_id}, {"$set": sample_dict})


def update_batch(adapter: StatinaAdapter, batch: Batch) -> None:
    """Update a batch object in the database"""
    batch_dict: dict = batch.dict(exclude_none=True)
    batch_id = batch.batch_id
    LOG.info("Updating batch %s", batch_id)
    adapter.batch_collection.update_one({"batch_id": batch_id}, {"$set": batch_dict})


def update_user(adapter: StatinaAdapter, user: User) -> None:
    """Update a user object in the database"""

    user_dict: dict = user.dict(exclude_none=True)
    email = user.email
    LOG.info(f"Updating user {email}")
    adapter.user_collection.update_one({"email": email}, {"$set": user_dict})



import logging

from mongo_adapter import MongoAdapter
from pymongo.collection import Collection
from pymongo.database import Database

LOG = logging.getLogger(__name__)


class StatinaAdapter(MongoAdapter):
    def setup(self, db_name: str):
        """Setup connection to a database"""

        if self.client is None:
            raise SyntaxError("No client is available")
        self.db: Database = self.client[db_name]
        self.db_name: str = db_name
        self.sample_collection: Collection = self.db.sample
        self.batch_collection: Collection = self.db.batch
        self.user_collection: Collection = self.db.user

        LOG.info("Use database %s.", db_name)



from typing import Dict, List

from statina.models.database import DataBaseSample
from statina.models.server.plots.coverage import CoveragePlotSampleData
from statina.models.server.sample import Sample, SampleWarning


def get_scatter_data_for_coverage_plot(
    samples: List[Sample],
) -> Dict["str", CoveragePlotSampleData]:
    """Coverage Ratio data for Coverage Plot.
    Only adding samples with a zscore warning"""

    data = {}
    for sample in samples:
        sample_warnings: SampleWarning = sample.warnings
        zscore_warnings = [
            sample_warnings.Zscore_13,
            sample_warnings.Zscore_18,
            sample_warnings.Zscore_21,
        ]
        if set(zscore_warnings) == {"default"}:
            continue

        x = []
        y = []
        for chromosome in range(1, 23):
            ratio = sample.dict().get(f"Chr{chromosome}_Ratio")
            if ratio is None:
                continue
            y.append(ratio)
            x.append(chromosome)
        data[sample.sample_id] = CoveragePlotSampleData(x_axis=x, y_axis=y)
    return data


def get_box_data_for_coverage_plot(samples: List[DataBaseSample]) -> Dict[int, List[float]]:
    """Coverage Ratio data for Coverage Plot."""

    data = {}
    for chromosome in range(1, 23):
        data[chromosome] = []
        for sample in samples:
            ratio = sample.dict().get(f"Chr{chromosome}_Ratio")
            if ratio is None:
                continue
            data[chromosome].append(ratio)
    return data



from typing import Dict, List

from fastapi import APIRouter, Depends, Request

from statina.adapter import StatinaAdapter
from statina.API.external.api.deps import get_current_user
from statina.API.external.constants import CHROM_ABNORM, STATUS_CLASSES, STATUS_COLORS
from statina.config import get_nipt_adapter, templates
from statina.crud.find import find
from statina.crud.find.plots.zscore_plot_data import (
    get_abn_for_samp_tris_plot,
    get_normal_for_samp_tris_plot,
    get_sample_for_samp_tris_plot,
)
from statina.models.database import Batch, DataBaseSample, User
from statina.models.server.plots.ncv import Zscore131821, ZscoreSamples
from statina.models.server.sample import Sample

router = APIRouter()


@router.get("/samples/")
def samples(
    request: Request,
    adapter: StatinaAdapter = Depends(get_nipt_adapter),
    user: User = Depends(get_current_user),
):
    """Get sample with id"""
    samples: List[DataBaseSample] = find.samples(adapter=adapter)
    return templates.TemplateResponse(
        "sample/samples.html",
        context=dict(
            request=request,
            current_user=user,
            sample_info=[Sample(**sample.dict()) for sample in samples],
            page_id="samples",
        ),
    )


@router.get("/samples/{sample_id}/")
def sample(
    request: Request,
    sample_id: str,
    adapter: StatinaAdapter = Depends(get_nipt_adapter),
    user: User = Depends(get_current_user),
):
    """Get sample with id"""

    sample: DataBaseSample = find.sample(sample_id=sample_id, adapter=adapter)
    batch: Batch = find.batch(batch_id=sample.batch_id, adapter=adapter)
    return templates.TemplateResponse(
        "sample/sample.html",
        context=dict(
            request=request,
            current_user=user,
            chrom_abnorm=CHROM_ABNORM,
            sample=Sample(**sample.dict()),
            status_classes=STATUS_CLASSES,
            batch=batch,
            page_id="sample",
        ),
    )


@router.post("/samples/{sample_id}/")
def sample(
    request: Request,
    sample_id: str,
    adapter: StatinaAdapter = Depends(get_nipt_adapter),
    user: User = Depends(get_current_user),
):
    """Post sample with id"""

    sample: DataBaseSample = find.sample(sample_id=sample_id, adapter=adapter)
    batch: Batch = find.batch(batch_id=sample.batch_id, adapter=adapter)
    return templates.TemplateResponse(
        "sample/sample.html",
        context=dict(
            request=request,
            current_user=user,
            chrom_abnorm=CHROM_ABNORM,
            sample=Sample(**sample.dict()),
            status_classes=STATUS_CLASSES,
            batch=batch,
            page_id="sample",
        ),
    )


@router.get("/samples/{sample_id}/tris")
def sample_tris(
    request: Request,
    sample_id: str,
    adapter: StatinaAdapter = Depends(get_nipt_adapter),
    user: User = Depends(get_current_user),
):
    """Sample view with trisomi plot."""
    sample: DataBaseSample = find.sample(sample_id=sample_id, adapter=adapter)
    batch: Batch = find.batch(batch_id=sample.batch_id, adapter=adapter)
    abnormal_data: Dict[str, ZscoreSamples] = get_abn_for_samp_tris_plot(adapter=adapter)
    normal_data: Zscore131821 = get_normal_for_samp_tris_plot(adapter=adapter)
    sample_data: ZscoreSamples = get_sample_for_samp_tris_plot(sample)
    return templates.TemplateResponse(
        "sample/sample_tris.html",
        context=dict(
            request=request,
            current_user=user,
            normal_data=normal_data.dict(exclude_none=True, by_alias=True),
            abnormal_data=abnormal_data,
            sample_data=sample_data,
            sample=Sample(**sample.dict()),
            batch=batch,
            status_colors=STATUS_COLORS,
            page_id="sample_tris",
        ),
    )


@router.post("/samples/{sample_id}/tris")
def sample_tris(
    request: Request,
    sample_id: str,
    adapter: StatinaAdapter = Depends(get_nipt_adapter),
    user: User = Depends(get_current_user),
):
    """Sample view with trisomi plot."""
    sample: DataBaseSample = find.sample(sample_id=sample_id, adapter=adapter)
    batch: Batch = find.batch(batch_id=sample.batch_id, adapter=adapter)
    abnormal_data: Dict[str, ZscoreSamples] = get_abn_for_samp_tris_plot(adapter=adapter)
    normal_data: Zscore131821 = get_normal_for_samp_tris_plot(adapter=adapter)
    sample_data: ZscoreSamples = get_sample_for_samp_tris_plot(sample)
    return templates.TemplateResponse(
        "sample/sample_tris.html",
        context=dict(
            request=request,
            current_user=user,
            normal_data=normal_data,
            abnormal_data=abnormal_data,
            sample_data=sample_data,
            sample=Sample(**sample.dict()),
            batch=batch,
            status_colors=STATUS_COLORS,
            page_id="sample_tris",
        ),
    )



from typing import Iterable, List, Optional

from pydantic import parse_obj_as

from statina.adapter import StatinaAdapter
from statina.models.database import Batch, DataBaseSample, User


def users(adapter: StatinaAdapter) -> List[User]:
    """Return all users from the batch collection"""

    users: Iterable[dict] = adapter.user_collection.find()
    return parse_obj_as(List[User], list(users))


def user(
    adapter: StatinaAdapter, email: Optional[str] = None, user_name: Optional[str] = None
) -> Optional[User]:
    """Find user from user collection"""
    if email:
        raw_user: dict = adapter.user_collection.find_one({"email": email})  # ????? wierd
    elif user_name:
        raw_user: dict = adapter.user_collection.find_one({"username": user_name})
    else:
        raise SyntaxError("Have to use email or user_name")
    if not raw_user:
        return None
    return User(**raw_user)


def samples(adapter: StatinaAdapter) -> List[DataBaseSample]:
    """Return all batches from the batch collection"""

    raw_samples: Iterable[dict] = adapter.sample_collection.find()
    return parse_obj_as(List[DataBaseSample], list(raw_samples))


def sample(adapter: StatinaAdapter, sample_id: str) -> Optional[DataBaseSample]:
    """Find one sample from the sample collection"""

    raw_sample = adapter.sample_collection.find_one({"sample_id": sample_id})
    if not raw_sample:
        return None

    return DataBaseSample(**raw_sample)


def batch(adapter: StatinaAdapter, batch_id: str) -> Optional[Batch]:
    """Find one batch from the batch collection"""

    raw_batch: dict = adapter.batch_collection.find_one({"batch_id": batch_id})
    if not raw_batch:
        return None
    return Batch(**raw_batch)


def batches(adapter: StatinaAdapter) -> List[Batch]:
    """Return all batches from the batch collection"""

    raw_batches: Iterable[dict] = adapter.batch_collection.find()
    return parse_obj_as(List[Batch], list(raw_batches))


def sample_aggregate(adapter: StatinaAdapter, pipe: list) -> list:
    """Aggregates a query pipeline on the sample collection"""

    return list(adapter.sample_collection.aggregate(pipe))


def batch_aggregate(adapter: StatinaAdapter, pipe: list) -> List[Batch]:
    """Aggregates a query pipeline on the sample collection"""

    return list(adapter.batch_collection.aggregate(pipe))


def batch_samples(adapter: StatinaAdapter, batch_id: str) -> List[DataBaseSample]:
    """All samples within the batch"""

    raw_samples: Iterable[dict] = adapter.sample_collection.find({"batch_id": batch_id})
    return parse_obj_as(List[DataBaseSample], list(raw_samples))



from pathlib import Path

from fastapi import APIRouter, Depends
from fastapi.responses import FileResponse, RedirectResponse
from starlette.requests import Request
from starlette.responses import StreamingResponse

from statina.API.external.constants import COLORS
from statina.adapter.plugin import StatinaAdapter
from statina.API.external.api.deps import get_current_user
from statina.config import get_nipt_adapter, templates
from statina.crud.find import find
from statina.crud.find.plots import fetal_fraction_plot_data as get_fetal_fraction
from statina.models.server.plots.fetal_fraction import (
    FetalFractionSamples,
    FetalFractionControlAbNormal,
)
from statina.models.server.plots.fetal_fraction_sex import SexChromosomeThresholds
from statina.models.server.sample import Sample
from zipfile import ZIP_DEFLATED, ZipFile
import io
from os import PathLike
from typing import Union, List
from statina.models.database import DataBaseSample, User, Batch
from statina.parse.batch import validate_file_path

router = APIRouter()


def zip_dir(source_dir: Union[str, PathLike]) -> io.BytesIO:
    """Function for zipping"""
    src_path = Path(source_dir).expanduser().resolve(strict=True)
    file_obj = io.BytesIO()
    file_obj.seek(0)
    with ZipFile(file=file_obj, mode="a", compression=ZIP_DEFLATED, compresslevel=9) as zf:
        for file in src_path.iterdir():
            zf.write(filename=file.as_posix(), arcname=file.name)
    file_obj.seek(0)
    return file_obj


@router.get("/batch_download/{batch_id}/{file_id}/{file_name}")
def batch_download(
    request: Request,
    batch_id: str,
    file_id: str,
    file_name: str,
    adapter: StatinaAdapter = Depends(get_nipt_adapter),
    user: User = Depends(get_current_user),
):
    """View for batch downloads"""
    batch: dict = find.batch(adapter=adapter, batch_id=batch_id).dict()
    file_path = batch.get(file_id)
    if not validate_file_path(file_path):
        return RedirectResponse(request.headers.get("referer"))

    path = Path(file_path)
    if path.is_dir():
        file_obj = zip_dir(source_dir=file_path)
        return StreamingResponse(file_obj, media_type="application/text")

    return FileResponse(
        str(path.absolute()), media_type="application/octet-stream", filename=path.name
    )


@router.get("/sample_download/{sample_id}/{file_id}")
def sample_download(
    request: Request,
    sample_id: str,
    file_id: str,
    adapter: StatinaAdapter = Depends(get_nipt_adapter),
    user: User = Depends(get_current_user),
):
    """View for sample downloads"""

    sample: DataBaseSample = find.sample(adapter=adapter, sample_id=sample_id)
    file_path = sample.dict().get(file_id)
    if not validate_file_path(file_path):
        # warn file missing!
        return RedirectResponse(request.headers.get("referer"))

    file = Path(file_path)
    return FileResponse(
        str(file.absolute()), media_type="application/octet-stream", filename=file.name
    )


@router.get("/batches/{batch_id}/report/{file_name}")
def report(
    request: Request,
    batch_id: str,
    file_name: str,
    adapter: StatinaAdapter = Depends(get_nipt_adapter),
    user: User = Depends(get_current_user),
):
    """Report view, collecting all tables and plots from one batch."""

    samples: List[DataBaseSample] = find.batch_samples(batch_id=batch_id, adapter=adapter)
    batch: Batch = find.batch(batch_id=batch_id, adapter=adapter)

    cases = get_fetal_fraction.samples(adapter=adapter, batch_id=batch_id)
    control: FetalFractionSamples = get_fetal_fraction.samples(
        batch_id=batch_id, adapter=adapter, control_samples=True
    )
    abnormal: FetalFractionControlAbNormal = get_fetal_fraction.control_abnormal(adapter)
    abnormal_dict = abnormal.dict(
        exclude_none=True,
        exclude={
            "X0": {"status_data_"},
            "XXX": {"status_data_"},
            "XXY": {"status_data_"},
            "XYY": {"status_data_"},
        },
    )

    x_max = max(control.FFX + cases.FFX) + 1
    x_min = min(control.FFX + cases.FFX) - 1

    sex_thresholds = SexChromosomeThresholds(x_min=x_min, x_max=x_max)

    template = templates.get_template("batch/report.html")
    output_from_parsed_template = template.render(
        # common
        request=request,
        current_user=user,
        batch=find.batch(batch_id=batch_id, adapter=adapter),
        # Fetal Fraction  XY
        sex_thresholds={
            "XY_fetal_fraction_y": sex_thresholds.XY_fetal_fraction_y(),
            "XX_lower": sex_thresholds.XX_lower(),
            "XX_upper": sex_thresholds.XX_upper(),
            "XY_upper": sex_thresholds.XY_upper(),
            "XY_lower": sex_thresholds.XY_lower(),
            "XXY": sex_thresholds.XXY(),
        },
        control=control,
        abnormal=abnormal_dict,
        cases=cases,
        colors=COLORS,
        max_x=x_max,
        min_x=x_min,
        # table
        sample_info=[Sample(**sample.dict()) for sample in samples],
    )

    in_mem_file = io.StringIO()
    in_mem_file.seek(0)

    in_mem_file.write(output_from_parsed_template)
    in_mem_file.seek(0)
    return StreamingResponse(in_mem_file, media_type="application/text")



from fastapi import APIRouter, Request

from statina.config import templates

router = APIRouter()


@router.post("/")
def index(request: Request):
    """Log in view."""
    return templates.TemplateResponse(
        "index.html",
        context={
            "request": request,
            "current_user": "",
            "info_type": request.cookies.get("info_type"),
            "user_info": request.cookies.get("user_info"),
        },
    )


@router.get("/")
def index(request: Request):
    """Log in view."""
    return templates.TemplateResponse(
        "index.html",
        context={
            "request": request,
            "current_user": "",
            "info_type": request.cookies.get("info_type"),
            "user_info": request.cookies.get("user_info"),
        },
    )



from typing import Optional

from statina.adapter import StatinaAdapter
from statina.API.external.constants import CHROM_ABNORM, SEX_CHROM_ABNORM
from statina.crud.find import find
from statina.models.server.plots.fetal_fraction import (
    AbNormalityClasses,
    FetalFractionControlAbNormal,
    FetalFractionSamples,
)


def samples(
    adapter: StatinaAdapter, batch_id: Optional[str] = None, control_samples: Optional[bool] = False
) -> FetalFractionSamples:
    """Samples for fetal fraction plot"""

    match = {
        "$match": {
            "FF_Formatted": {"$ne": "None", "$exists": True},
            "FFY": {"$ne": "None", "$exists": True},
            "FFX": {"$ne": "None", "$exists": True},
        }
    }
    group = {
        "$group": {
            "_id": "None",
            "FFY": {"$push": "$FFY"},
            "FFX": {"$push": "$FFX"},
            "FF": {"$push": "$FF_Formatted"},
            "names": {"$push": "$sample_id"},
            "count": {"$sum": 1},
        }
    }

    if control_samples:
        match["$match"]["batch_id"] = {"$ne": batch_id}
        match["$match"]["include"] = {"$eq": True}
        for abn in SEX_CHROM_ABNORM:
            match["$match"][f"status_{abn}"] = {"$eq": "Normal"}
    else:
        match["$match"]["batch_id"] = {"$eq": batch_id}

    relevant_aggregation_data = list(find.sample_aggregate(pipe=[match, group], adapter=adapter))[0]
    return FetalFractionSamples(**relevant_aggregation_data)


def control_abnormal(adapter: StatinaAdapter) -> FetalFractionControlAbNormal:
    """Abnormal Control Samples for fetal_fraction_XY plot"""

    plot_data = {}
    for abn in CHROM_ABNORM:
        plot_data[abn] = {}
        pipe = [
            {
                "$match": {
                    f"status_{abn}": {"$ne": "Normal", "$exists": True},
                    "include": {"$eq": True},
                }
            },
            {
                "$group": {
                    "_id": {f"status_{abn}": f"$status_{abn}"},
                    "FFX": {"$push": "$FFX"},
                    "FFY": {"$push": "$FFY"},
                    "names": {"$push": "$sample_id"},
                    "count": {"$sum": 1},
                }
            },
        ]
        statuses = {}
        for status_dict in find.sample_aggregate(pipe=pipe, adapter=adapter):
            status: str = status_dict["_id"][f"status_{abn}"]
            statuses[status] = status_dict

        plot_data[abn] = AbNormalityClasses(status_data_=statuses)
    return FetalFractionControlAbNormal(**plot_data)



from typing import Optional

from pydantic import BaseModel, Field


class Batch(BaseModel):
    batch_id: str = Field(..., alias="SampleProject")
    result_file: Optional[str]
    multiqc_report: Optional[str]
    segmental_calls: Optional[str]
    Flowcell: Optional[str]
    SequencingDate: Optional[str]
    Median_13: Optional[float]
    Median_18: Optional[float]
    Median_21: Optional[float]
    Median_X: Optional[float]
    Median_Y: Optional[float]
    Stdev_13: Optional[float]
    Stdev_18: Optional[float]
    Stdev_21: Optional[float]
    Stdev_X: Optional[float]
    Stdev_Y: Optional[float]
    comment: Optional[str] = ""

    class Config:
        validate_assignment = True
        allow_population_by_field_name = True



from typing import List

from pydantic import BaseModel, validator


class CoveragePlotSampleData(BaseModel):
    """Sample data for coverage plot"""

    y_axis: List[float]
    x_axis: List[int]

    @validator("x_axis")
    def check_list_lengths(cls, v, values: dict) -> List[int]:
        if len(v) != len(values["y_axis"]):
            raise ValueError("Axis lengths do not match")
        return v



from typing import List, Optional

import numpy as np
from pydantic import BaseModel, validator, root_validator


class ThresholdLine(BaseModel):
    x: List[float]
    y: List[float]
    text: str
    text_position: Optional[List[float]]

    @validator("text_position", always=True)
    def set_text_position(cls, v, values: dict) -> List[np.ndarray]:
        """ """

        return [np.mean(values["x"]), np.mean(values["y"])]


class SexChromosomeThresholds(BaseModel):
    y_min: float = -1
    y_max: float = 20
    x_min: float
    x_max: float
    xx_lower: float = 2
    xx_upper: float = 4
    xy_lowest: float = 0.5
    XXY: Optional[ThresholdLine]
    XY_upper: Optional[ThresholdLine]
    XY_lower: Optional[ThresholdLine]
    XY_horisontal: Optional[ThresholdLine]
    XX_upper: Optional[ThresholdLine]
    XX_lower: Optional[ThresholdLine]

    @classmethod
    def straight_line_get_y(cls, x: float, m: float, k: float) -> float:
        return k * x + m

    @classmethod
    def xy_lower_get_y(cls, x: float) -> float:
        k = 1.51
        m = 0.5 - k * 4
        return k * x + m

    @classmethod
    def xy_upper_get_y(cls, x: float) -> float:
        k = 1.49
        m = 0.5 - k * 2
        return k * x + m

    @classmethod
    def xy_upper_get_x(cls, y: float) -> float:
        k = 1.49
        m = 0.5 - k * 2
        return (y - m) / k

    @validator("XXY", always=True)
    def set_XXY(cls, v, values: dict) -> ThresholdLine:
        """ """
        x = values["xx_upper"]

        return ThresholdLine(x=[x, x], y=[values["y_max"], cls.xy_upper_get_y(x=x)], text=f"x={x}")

    @validator("XY_lower", always=True)
    def set_XY_upper(cls, v, values: dict) -> ThresholdLine:
        """ """
        return ThresholdLine(
            x=[values["x_max"], values["xx_upper"]],
            y=[
                cls.xy_lower_get_y(x=values["x_max"]),
                cls.xy_lower_get_y(x=values["xx_upper"]),
            ],
            text=f"hej",
        )

    @validator("XY_upper", always=True)
    def set_XY_lower(cls, v, values: dict) -> ThresholdLine:
        """ """
        return ThresholdLine(
            x=[values["x_max"], values["xx_lower"]],
            y=[
                cls.xy_upper_get_y(x=values["x_max"]),
                cls.xy_upper_get_y(x=values["xx_lower"]),
            ],
            text=f"hej",
        )

    @validator("XY_horisontal", always=True)
    def set_XY_horisontal(cls, v, values: dict) -> ThresholdLine:
        """ """
        y = values["xy_lowest"]
        return ThresholdLine(x=[values["x_min"], values["x_max"]], y=[y, y], text=f"y={y}")

    @validator("XX_upper", always=True)
    def set_XX_upper(cls, v, values: dict) -> ThresholdLine:
        """ """
        x = values["xx_upper"]

        return ThresholdLine(x=[x, x], y=[values["y_min"], values["xy_lowest"]], text=f"x={x}")

    @validator("XX_lower", always=True)
    def set_XX_lower(cls, v, values: dict) -> ThresholdLine:
        """ """
        x = values["xx_lower"]

        return ThresholdLine(x=[x, x], y=[values["y_min"], values["xy_lowest"]], text=f"x={x}")



import csv
import logging
from pathlib import Path
from typing import Dict, List, Optional

from pydantic import parse_obj_as

from statina.models.database import Batch, DataBaseSample

LOG = logging.getLogger(__name__)


def parse_segmental_calls(segmental_calls_path: Optional[str]) -> dict:
    """Builds a dict with segmental calls bed files.
    key: sample ids
    value: bed file path"""

    segmental_calls = {}
    if not validate_file_path(segmental_calls_path):
        LOG.warning("Segmental Calls file path missing.")
        return segmental_calls

    segmental_calls_dir = Path(segmental_calls_path)
    for file in segmental_calls_dir.iterdir():
        if file.suffix == ".bed":
            sample_id = file.name.split(".")[0]
            segmental_calls[sample_id] = str(file.absolute())

    return segmental_calls


def validate_file_path(file_path: Optional[str]) -> bool:
    """File path validation"""

    if not file_path:
        return False
    file = Path(file_path)
    return file.exists()


def convert_empty_str_to_none(data: dict) -> dict:
    """Convert all values that are empty string to None in a dict"""

    for key, value in data.items():
        if not value:
            data[key] = None
    return data


def parse_csv(infile: Path) -> List[Dict[str, str]]:
    with open(infile, "r") as csv_file:
        entries = [convert_empty_str_to_none(entry) for entry in csv.DictReader(csv_file)]
    return entries


def get_samples(nipt_results_path: Path) -> List[DataBaseSample]:
    """Parse NIPT result file into samples"""

    return parse_obj_as(List[DataBaseSample], parse_csv(nipt_results_path))


def get_batch(nipt_results_path: Path) -> Batch:
    """Parse NIPT result file and create a batch object from the first sample information"""

    sample_data: List[dict] = parse_csv(nipt_results_path)

    return Batch.parse_obj(sample_data[0])



import logging
from pathlib import Path
from typing import List

import click
from pymongo import MongoClient

from statina.adapter import StatinaAdapter
from statina.config import settings
from statina.crud import find, insert
from statina.exeptions import InsertError
from statina.models.database import Batch, DataBaseSample
from statina.models.server.load import BatchRequestBody, UserRequestBody
from statina.parse.batch import get_batch, get_samples

LOG = logging.getLogger(__name__)


@click.group(name="load")
@click.pass_obj
def load_commands(context: dict):
    client = MongoClient(settings.db_uri)
    context["adapter"] = StatinaAdapter(client, db_name=settings.db_name)
    LOG.info("Connected to %s", settings.db_name)


@load_commands.command(name="batch")
@click.option("--result-file", type=click.Path(exists=True, dir_okay=False), required=True)
@click.option("--multiqc-report", type=click.Path(dir_okay=False))
@click.option("--segmental-calls", type=click.Path(dir_okay=False))
@click.pass_obj
def load_batch(
    context: dict, result_file: click.Path, multiqc_report: click.Path, segmental_calls: click.Path
):
    """Load fluffy result into database"""
    batch_files: BatchRequestBody = BatchRequestBody(
        result_file=str(result_file),
        multiqc_report=str(multiqc_report),
        segmental_calls=str(segmental_calls),
    )
    adapter: StatinaAdapter = context["adapter"]
    nipt_results = Path(str(result_file))
    samples: List[DataBaseSample] = get_samples(nipt_results)
    batch: Batch = get_batch(nipt_results)
    if find.batch(adapter=adapter, batch_id=batch.batch_id):
        return "batch already in database"
    insert.insert_batch(adapter=adapter, batch=batch, batch_files=batch_files)
    insert.insert_samples(
        adapter=adapter, samples=samples, segmental_calls=batch_files.segmental_calls
    )


@load_commands.command(name="user")
@click.option("--email", required=True)
@click.option("--user-name", required=True)
@click.option("--role", type=click.Choice(["RW", "R", "admin"]), default="RW", show_default=True)
@click.pass_obj
def load_user(context: dict, email: str, user_name: str, role: str):
    """Add a user to the database"""
    user: UserRequestBody = UserRequestBody(email=email, username=user_name, role=role)
    adapter: StatinaAdapter = context["adapter"]

    try:
        insert.insert_user(adapter=adapter, user=user)
    except InsertError as err:
        LOG.warning(err)
        raise click.Abort



from setuptools import find_packages, setup


try:
    with open("requirements.txt", "r") as f:
        install_requires = [x.strip() for x in f.readlines()]
except IOError:
    install_requires = []


with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="statina",
    version="1.2.7",
    description="NIPT data storage and visualisation",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Clinical-Genomics/statina",
    author="Maya Brandi",
    author_email="maya.brandi@scilifelab.se",
    install_requires=install_requires,
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    entry_points={
        "console_scripts": ["statina=statina.commands:cli"],
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
)



"""
Module with CLI commands for Statina

The CLI is intended for development/testing purpose only. To run in a production setting please refer to documentation
for suggestions how.

"""
import logging

import click
import pkg_resources
import uvicorn

# Get version and doc decorator
from statina import __version__
from statina.commands.load_commands import load_commands
from statina.config import settings

LOG_LEVELS = ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]
LOG = logging.getLogger(__name__)

ENV_FILE = pkg_resources.resource_filename("statina", ".env")


@click.version_option(__version__)
@click.group()
@click.pass_context
def cli(context: click.Context):
    """Main entry point"""
    logging.basicConfig(level=logging.INFO)
    context.obj = {}


@cli.command(name="serve")
@click.option(
    "--api", default="external", type=click.Choice(["external", "internal"]), show_default=True
)
@click.option("--reload", is_flag=True)
def serve_command(reload: bool, api: str):
    """Serve the Statina app for testing purpose.

    This command will serve the user interface (external) as default
    """
    app = f"statina.main:{api}_app"
    LOG.info("Running %s api on host:%s and port:%s", api, settings.host, settings.port)
    uvicorn.run(app=app, host=settings.host, port=settings.port, reload=reload)


cli.add_command(load_commands)



from pydantic import BaseModel, EmailStr, validator

from statina.exeptions import MissMatchingPasswordError
from statina.config import EmailSettings


class NewUser(BaseModel):
    email: EmailStr
    username: str
    password_repeated: str
    password: str

    @validator("password", always=True)
    def validate_password(cls, v, values: dict) -> str:
        if v != values["password_repeated"]:
            raise MissMatchingPasswordError
        return v


class NewUserRequestEmail(EmailSettings):
    subject: str = "Statina User Request"
    user_email: EmailStr
    message: str

    @validator("message", always=True)
    def make_message(cls, v, values: dict) -> str:
        return f"User request from {v}. Give the user proper credentials and send an email to {values['user_email']} when done."



from dataclasses import dataclass
from typing import List, Optional

import numpy as np
from pydantic import BaseModel, validator

from statina.API.external.constants import FF_TRESHOLDS


class ThresholdLine(BaseModel):
    """Model for treshold lines for plots."""

    x: List[float]
    y: List[float]
    text: str
    text_position: Optional[List[float]]

    @validator("text_position", always=True)
    def set_text_position(cls, v, values: dict) -> List[np.ndarray]:
        """ """
        return [np.mean(values["x"]), np.mean(values["y"])]


def x_get_y(x: float, k: float, m: float) -> float:
    return k * x + m


def y_get_x(y: float, k: float, m: float) -> float:
    return (y - m) / k


class SexChromosomeThresholds:
    """Threshold lines for the Fetal Fraction XY plot"""

    y_axis_min: float = FF_TRESHOLDS["y_axis_min"]
    y_axis_max: float = FF_TRESHOLDS["y_axis_max"]
    xx_lower: float = FF_TRESHOLDS["fetal_fraction_XXX"]
    xx_upper: float = FF_TRESHOLDS["fetal_fraction_X0"]
    xy_lowest: float = FF_TRESHOLDS["fetal_fraction_y_min"]
    k_upper: float = FF_TRESHOLDS["k_upper"]
    k_lower: float = FF_TRESHOLDS["k_lower"]
    m_lower: float = FF_TRESHOLDS["m_lower"]
    m_upper: float = FF_TRESHOLDS["m_upper"]

    def __init__(self, x_min, x_max):
        self.x_min: float = x_min
        self.x_max: float = x_max

    def XXY(self) -> ThresholdLine:
        """Returning a threshold line to separate XYY from XXY"""
        x = self.xx_upper
        return ThresholdLine(
            x=[x, x],
            y=[
                x_get_y(x=x, k=self.k_upper, m=self.m_upper),
                x_get_y(x=self.x_max, k=self.k_upper, m=self.m_upper),
            ],
            text=f"x = {x}",
        )

    def XY_lower(self) -> ThresholdLine:
        """Returning a threshold line to separate XY from other"""
        return ThresholdLine(
            x=[y_get_x(y=self.xy_lowest, k=self.k_lower, m=self.m_lower), self.x_max],
            y=[
                self.xy_lowest,
                x_get_y(x=self.x_max, k=self.k_lower, m=self.m_lower),
            ],
            text=f"y = {self.k_lower}x + {self.m_lower}",
        )

    def XY_upper(self) -> ThresholdLine:
        """Returning a threshold line to separate XY from XXY and XYY"""
        return ThresholdLine(
            x=[y_get_x(y=self.xy_lowest, k=self.k_upper, m=self.m_upper), self.x_max],
            y=[
                self.xy_lowest,
                x_get_y(x=self.x_max, k=self.k_upper, m=self.m_upper),
            ],
            text=f"y = {self.k_upper}x + {self.m_upper}",
        )

    def XY_fetal_fraction_y(self) -> ThresholdLine:
        """Returning a threshold line to separate XY from XX"""
        y = self.xy_lowest
        return ThresholdLine(x=[self.x_min, self.x_max], y=[y, y], text=f"y = {y}")

    def XX_upper(self) -> ThresholdLine:
        """Returning a threshold line to separate XX from XXX"""
        x = self.xx_upper

        return ThresholdLine(x=[x, x], y=[self.y_axis_min, self.xy_lowest], text=f"x = {x}")

    def XX_lower(self) -> ThresholdLine:
        """Returning a threshold line to separate XX from X0"""
        x = self.xx_lower

        return ThresholdLine(x=[x, x], y=[self.y_axis_min, self.xy_lowest], text=f"x = {x}")



from copy import deepcopy
from typing import Dict, Optional

from statina.adapter import StatinaAdapter
from statina.crud.find import find
from statina.models.database import DataBaseSample
from statina.models.server.plots.ncv import Zscore131821, ZscoreSamples


def get_tris_control_abnormal(adapter: StatinaAdapter, chr, x_axis) -> Dict[str, ZscoreSamples]:
    """Abnormal Control Samples for trisomi plots"""

    plot_data = {}

    pipe = [
        {
            "$match": {
                f"status_{chr}": {"$ne": "Normal", "$exists": "True"},
                "include": {"$eq": True},
            }
        },
        {
            "$group": {
                "_id": {f"status_{chr}": f"$status_{chr}"},
                "ncv_values": {"$push": f"$Zscore_{chr}"},
                "names": {"$push": "$sample_id"},
                "count": {"$sum": 1},
            }
        },
    ]

    for status_dict in find.sample_aggregate(pipe=pipe, adapter=adapter):
        status = status_dict["_id"][f"status_{chr}"]
        plot_data[status] = ZscoreSamples(
            ncv_values=[value for value in status_dict.get("ncv_values")],
            names=status_dict.get("names"),
            count=status_dict.get("count"),
            x_axis=[x_axis] * status_dict.get("count"),
        )

    return plot_data


def get_abn_for_samp_tris_plot(adapter: StatinaAdapter) -> Dict[str, ZscoreSamples]:
    """Format abnormal Control Samples for Sample trisomi plot"""

    plot_data = {}

    for x_axis, abn in enumerate(["13", "18", "21"], start=1):
        tris_control_abnormal: Dict[str, ZscoreSamples] = get_tris_control_abnormal(
            adapter, abn, x_axis
        )
        for status, data in tris_control_abnormal.items():
            if status not in plot_data:
                plot_data[status] = deepcopy(data)
                continue
            plot_data[status] = ZscoreSamples(
                ncv_values=plot_data[status].ncv_values + data.ncv_values,
                names=plot_data[status].names + data.names,
                count=plot_data[status].count + data.count,
                x_axis=plot_data[status].x_axis + data.x_axis,
            )
    return plot_data


def get_tris_control_normal(
    adapter: StatinaAdapter, chr: str, x_axis: Optional[int] = None
) -> ZscoreSamples:
    """Normal Control Samples for trisomi plots"""

    pipe = [
        {"$match": {f"status_{chr}": {"$eq": "Normal"}, "include": {"$eq": True}}},
        {
            "$group": {
                "_id": {f"status_{chr}": f"$status_{chr}"},
                "ncv_values": {"$push": f"$Zscore_{chr}"},
                "names": {"$push": "$sample_id"},
                "count": {"$sum": 1},
            }
        },
    ]
    if not list(find.sample_aggregate(pipe=pipe, adapter=adapter)):
        return {}
    data = find.sample_aggregate(pipe=pipe, adapter=adapter)[0]
    if x_axis:
        data["x_axis"] = [x_axis] * data.get("count")

    return ZscoreSamples(**data)


def get_normal_for_samp_tris_plot(adapter: StatinaAdapter) -> Zscore131821:
    """Format normal Control Samples for Sample trisomi plot"""

    return Zscore131821(
        chr_13=get_tris_control_normal(adapter=adapter, chr="13", x_axis=1),
        chr_18=get_tris_control_normal(adapter=adapter, chr="18", x_axis=2),
        chr_21=get_tris_control_normal(adapter=adapter, chr="21", x_axis=3),
    )


def get_abnormal_for_samp_tris_plot(adapter: StatinaAdapter) -> dict:
    """Format normal Control Samples for Sample trisomi plot"""

    return {
        "13": get_tris_control_abnormal(adapter=adapter, chr="13", x_axis=0),
        "18": get_tris_control_abnormal(adapter=adapter, chr="18", x_axis=0),
        "21": get_tris_control_abnormal(adapter=adapter, chr="21", x_axis=0),
    }


def get_samples_for_samp_tris_plot(adapter: StatinaAdapter, batch_id: str) -> Zscore131821:
    return Zscore131821(
        chr_13=get_tris_samples(adapter=adapter, chr="13", batch_id=batch_id),
        chr_18=get_tris_samples(adapter=adapter, chr="18", batch_id=batch_id),
        chr_21=get_tris_samples(adapter=adapter, chr="21", batch_id=batch_id),
    )


def get_tris_samples(adapter: StatinaAdapter, chr, batch_id: str) -> ZscoreSamples:
    """Cases for trisomi plots."""

    pipe = [
        {"$match": {"batch_id": {"$eq": batch_id}}},
        {
            "$group": {
                "_id": {"batch": "$batch_id"},
                "ncv_values": {"$push": f"$Zscore_{chr}"},
                "names": {"$push": "$sample_id"},
                "count": {"$sum": 1},
            }
        },
    ]

    if not list(find.sample_aggregate(pipe=pipe, adapter=adapter)):
        return {}

    data = list(find.sample_aggregate(pipe=pipe, adapter=adapter))[0]
    data["x_axis"] = list(range(1, data.get("count") + 1))

    return ZscoreSamples(**data)


def get_sample_for_samp_tris_plot(sample: DataBaseSample) -> ZscoreSamples:
    """Case data for Sample trisomi plot"""
    return ZscoreSamples(
        ncv_values=[sample.Zscore_13, sample.Zscore_18, sample.Zscore_21],
        names=[sample.sample_id, sample.sample_id, sample.sample_id],
        x_axis=[1, 2, 3],
    )



from typing import Literal, Optional

from pydantic import BaseModel, Field

status_options = Literal[
    "Probable",
    "Suspected",
    "False Positive",
    "Verified",
    "False Negative",
    "Other",
    "Failed",
    "Normal",
]


class DataBaseSample(BaseModel):
    sample_id: str = Field(..., alias="SampleID")
    batch_id: str = Field(..., alias="SampleProject")
    SampleType: Optional[str] = ""
    Description: Optional[str]
    Flowcell: Optional[str]
    Index1: Optional[str]
    Index2: Optional[str]
    Library_nM: Optional[float]
    QCFlag: Optional[str] = ""
    Zscore_13: Optional[float]
    Zscore_18: Optional[float]
    Zscore_21: Optional[float]
    Zscore_X: Optional[float]
    MappedReads: Optional[int]
    GC_Dropout: Optional[float]
    AT_Dropout: Optional[float]
    Chr1_Ratio: Optional[float]
    Chr2_Ratio: Optional[float]
    Chr3_Ratio: Optional[float]
    Chr4_Ratio: Optional[float]
    Chr5_Ratio: Optional[float]
    Chr6_Ratio: Optional[float]
    Chr7_Ratio: Optional[float]
    Chr8_Ratio: Optional[float]
    Chr9_Ratio: Optional[float]
    Chr10_Ratio: Optional[float]
    Chr11_Ratio: Optional[float]
    Chr12_Ratio: Optional[float]
    Chr13_Ratio: Optional[float]
    Chr18_Ratio: Optional[float]
    Chr21_Ratio: Optional[float]
    ChrX_Ratio: Optional[float]
    ChrY_Ratio: Optional[float]
    Chr14_Ratio: Optional[float]
    Chr15_Ratio: Optional[float]
    Chr16_Ratio: Optional[float]
    Chr17_Ratio: Optional[float]
    Chr19_Ratio: Optional[float]
    Chr20_Ratio: Optional[float]
    Chr22_Ratio: Optional[float]
    Chr1: Optional[int]
    Chr2: Optional[int]
    Chr3: Optional[int]
    Chr4: Optional[int]
    Chr5: Optional[int]
    Chr6: Optional[int]
    Chr7: Optional[int]
    Chr8: Optional[int]
    Chr9: Optional[int]
    Chr10: Optional[int]
    Chr11: Optional[int]
    Chr12: Optional[int]
    Chr13: Optional[int]
    Chr14: Optional[int]
    Chr15: Optional[int]
    Chr16: Optional[int]
    Chr17: Optional[int]
    Chr18: Optional[int]
    Chr19: Optional[int]
    Chr20: Optional[int]
    Chr21: Optional[int]
    Chr22: Optional[int]
    ChrX: Optional[int]
    ChrY: Optional[int]
    FF_Formatted: Optional[float] = Field(..., alias="Fetal Fraction Preface")
    FFY: Optional[float]
    FFX: Optional[float]
    DuplicationRate: Optional[float]
    Bin2BinVariance: Optional[float]
    UnfilteredCNVcalls: Optional[int]
    CNVSegment: Optional[str]
    segmental_calls: Optional[str]
    include: Optional[bool]
    change_include_date: Optional[str] = ""
    comment: Optional[str] = ""
    status_13: Optional[status_options] = "Normal"
    status_18: Optional[status_options] = "Normal"
    status_21: Optional[status_options] = "Normal"
    status_X0: Optional[status_options] = "Normal"
    status_XXX: Optional[status_options] = "Normal"
    status_XXY: Optional[status_options] = "Normal"
    status_XYY: Optional[status_options] = "Normal"
    status_change_13: Optional[str] = ""
    status_change_18: Optional[str] = ""
    status_change_21: Optional[str] = ""
    status_change_X0: Optional[str] = ""
    status_change_XXX: Optional[str] = ""
    status_change_XXY: Optional[str] = ""
    status_change_XYY: Optional[str] = ""

    class Config:
        allow_population_by_field_name = True
